# Copyright (c) 2019-     Xilinx, Inc              (Giuseppe Franco)
# Copyright (c) 2018-     Xilinx, Inc              (Alessandro Pappalardo)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Optional

import brevitas.config as config
from brevitas.inject.defaults import Int8WeightPerTensorFloat
from .quant_scale_bias import QuantScaleBias
from .utils import mul_add_from_bn
from .quant_layer import WeightQuantType, BiasQuantType, ActQuantType


class BatchNorm2dToQuantScaleBias(QuantScaleBias):

    def __init__(
            self,
            num_features,
            eps: float = 1e-5,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs):
        super(BatchNorm2dToQuantScaleBias, self).__init__(
            num_features,
            bias=True,
            runtime_shape=(1, -1, 1, 1),
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self.eps = eps

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        weight_key = prefix + 'weight'
        bias_key = prefix + 'bias'
        running_mean_key = prefix + 'running_mean'
        running_var_key = prefix + 'running_var'
        num_batches_tracked_key = prefix + 'num_batches_tracked'
        if running_mean_key in state_dict and running_var_key in state_dict:
            weight_init, bias_init = mul_add_from_bn(
                bn_bias=state_dict[bias_key],
                bn_weight=state_dict[weight_key],
                bn_mean=state_dict[running_mean_key],
                bn_var=state_dict[running_var_key],
                bn_eps=self.eps)
            self.weight.data = weight_init
            self.bias.data = bias_init
            del state_dict[bias_key]
            del state_dict[weight_key]
            del state_dict[running_mean_key]
            del state_dict[running_var_key]
            del state_dict[num_batches_tracked_key]
        super(BatchNorm2dToQuantScaleBias, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        if config.IGNORE_MISSING_KEYS and bias_key in missing_keys:
            missing_keys.remove(bias_key)
        if config.IGNORE_MISSING_KEYS and weight_key in missing_keys:
            missing_keys.remove(weight_key)
        if num_batches_tracked_key in unexpected_keys:
            unexpected_keys.remove(num_batches_tracked_key)




import torch
import torch.nn as nn
from torch.autograd import Function

class squared_hinge_loss(Function):
    @staticmethod
    def forward(ctx, predictions, targets):
        ctx.save_for_backward(predictions, targets) 
        output = 1.-predictions.mul(targets)
        output[output.le(0.)] = 0.
        loss = torch.mean(output.mul(output))
        return loss 

    @staticmethod
    def backward(ctx, grad_output):
       predictions, targets = ctx.saved_tensors
       output=1.-predictions.mul(targets)
       output[output.le(0.)]=0.
       grad_output.resize_as_(predictions).copy_(targets).mul_(-2.).mul_(output)
       grad_output.mul_(output.ne(0).float())
       grad_output.div_(predictions.numel())
       return grad_output, None    

class SqrHingeLoss(nn.Module):
    # Squared Hinge Loss
    def __init__(self):
        super(SqrHingeLoss, self).__init__()
    
    def forward(self, input, target):
        return squared_hinge_loss.apply(input, target)


from typing import List
from copy import copy
from inspect import isclass

from torch import Tensor
from torch.nn import Parameter
from torch.nn import Module, Sequential, ModuleList, ModuleDict

from .module import CodegenModule, Index, Instruction
from .tracer.trace import Trace, TraceElem, FnType
from .utils import module_class_name, flatten


class ModuleGenerator(object):

    default_gen_module_allowlist = ['torch.nn', 'brevitas.nn']
    default_gen_module_blocklist = [Sequential, ModuleList, ModuleDict]

    def __init__(
            self,
            gen_module_allowlist=tuple(default_gen_module_allowlist),
            gen_module_blocklist=tuple(default_gen_module_blocklist)):
        self.gen_module_allowlist = list(gen_module_allowlist)
        self.gen_module_blocklist = list(gen_module_blocklist)
        self._generated_allowlist_modules = []
        self._module_name_id_dict = {}

    def _parent_in_gen_allowlist(self, trace_elem):
        # find parent that has to be preserved as-is, if any
        output = None
        # context is all parent modules, from the inner most to the outer most
        context = reversed(trace_elem.module_context_list[:-1])
        for c in context:
            for gma in self.gen_module_allowlist:
                if isinstance(gma, str) and module_class_name(c).startswith(gma):
                    if not self._is_module_in_gen_blocklist(c):
                        output = c
                elif isclass(gma) and isinstance(c, gma):
                    if not self._is_module_in_gen_blocklist(c):
                        output = c
        return output

    def _is_module_in_gen_list(self, module, gen_list):
        output = False
        for gm in gen_list:
            output |= isinstance(gm, str) and module_class_name(module).startswith(gm)
            output |= isclass(gm) and isinstance(module, gm)
        return output

    def _is_module_in_gen_blocklist(self, module):
        return self._is_module_in_gen_list(module, self.gen_module_blocklist)

    def _is_module_in_gen_allowlist(self, module):
        return self._is_module_in_gen_list(module, self.gen_module_allowlist)

    def _is_module_already_gen(self, module, trace_elem):
        is_already_gen = False
        # check that the module itself hasn't been generated yet
        # input and output have to be the same
        # otherwise it's a different invokation of the same module
        for m, i_list, o in self._generated_allowlist_modules:
            same_module = m is module
            same_input = all([i is mi for i, mi in zip(i_list, trace_elem.module_input_list)])
            same_output = o is trace_elem.module_output
            is_already_gen |= same_module and same_input and same_output
        # check that none of the parents has been already preserved as-is
        for c in trace_elem.module_context_list:  # TODO same input and output across parents
            for m in self._generated_allowlist_modules:
                is_already_gen |= c is m
        return is_already_gen

    def _add_module(self, module: Module, module_name, prefix_list, model: Module):
        supermodule = model
        for prefix in prefix_list:
            submodule_names, submodules = zip(*list(supermodule.named_modules()))
            if prefix in submodule_names:
                supermodule = supermodule._modules[prefix]
            else:
                submodule = Module()
                supermodule.add_module(prefix, submodule)
                supermodule = submodule
        supermodule.add_module(module_name, module)

    def _module_instruction(
            self, module: Module, trace_elem: TraceElem, trace: Trace, output_model: Module):
        module_name = trace_elem.prefix_list[-1]
        prefix_list = trace_elem.prefix_list[:-1]  # exclude module name
        self._add_module(module, module_name, prefix_list, output_model)
        iil = [trace.index_from_val(i) for i in trace_elem.module_input_list]
        module_output_index = trace.index_from_val(trace_elem.module_output)
        inst = Instruction(module_output_index, module, FnType.MODULE, iil, {}, trace_elem.prefix)
        return inst

    def _module_fn_instruction(self, trace_elem: TraceElem, output_model):
        module = trace_elem.fn
        module_name = trace_elem.module_fn_name
        self._add_module(module, module_name, trace_elem.prefix_list, output_model)
        return self._torch_fn_instruction(trace_elem)

    def _torch_fn_instruction(self, trace_elem: TraceElem):
        fn_args = trace_elem.fn_args_index
        fn_kwargs = trace_elem.fn_kwargs_index
        fn = trace_elem.fn
        fn_type = trace_elem.fn_type
        output_index = trace_elem.fn_out_index
        return Instruction(output_index, fn, fn_type, fn_args, fn_kwargs, trace_elem.prefix)

    def _tensor_fn_instruction(self, trace_elem: TraceElem, trace: Trace):
        fn = trace_elem.fn
        fn_type = trace_elem.fn_type
        fn_args = trace_elem.fn_args_index
        fn_kwargs = trace_elem.fn_kwargs_index
        output_index = trace_elem.fn_out_index
        return Instruction(output_index, fn, fn_type, fn_args, fn_kwargs, trace_elem.prefix)

    def _gen_schedule(self, trace: Trace, gen_model: Module):
        schedule = []
        for trace_elem in trace.trace_elem_list:
            module = trace_elem.module_context_list[-1]
            # if a parent is supposed to be preserved as is
            parent_in_allowlist = self._parent_in_gen_allowlist(trace_elem)
            # If the module wrapping this fn is supposed to be preserved as-is
            in_allowlist = self._is_module_in_gen_allowlist(module)
            in_allowlist &= not self._is_module_in_gen_blocklist(module)
            # if either the wrapping module or parents have been already preserved as-is
            m_already_gen = self._is_module_already_gen(module, trace_elem)
            inst = None
            if parent_in_allowlist is not None and not m_already_gen:
                ctx = (module, trace_elem.module_input_list, trace_elem.module_output)
                self._generated_allowlist_modules.append(ctx)
                inst = self._module_instruction(parent_in_allowlist, trace_elem, trace, gen_model)
            elif in_allowlist and not m_already_gen:
                ctx = (module, trace_elem.module_input_list, trace_elem.module_output)
                self._generated_allowlist_modules.append(ctx)
                inst = self._module_instruction(module, trace_elem, trace, gen_model)
            elif in_allowlist and m_already_gen:
                continue
            else:
                if trace_elem.fn_type == FnType.MODULE:
                    inst = self._module_fn_instruction(trace_elem, gen_model)
                elif trace_elem.fn_type == FnType.FUNCTION:
                    inst = self._torch_fn_instruction(trace_elem)
                elif trace_elem.fn_type == FnType.METHOD:
                    inst = self._tensor_fn_instruction(trace_elem, trace)
                elif trace_elem.fn_type == FnType.ATTRIBUTE:
                    inst = self._tensor_fn_instruction(trace_elem, trace)
                elif trace_elem.fn_type == FnType.SCRIPTMODULE:
                    inst = self._module_instruction(trace_elem.fn, trace_elem, trace, gen_model)
                assert inst is not None
            schedule.append(inst)
        return schedule

    # constants are all inputs not computed as outputs, excluding the model input
    def _gen_consts_params_buffers(self, schedule: List[Instruction], trace: Trace):
        output_index_list = [inst.output_index for inst in schedule]
        input_index_list = flatten([i for inst in schedule for i in inst.input_args_list])
        input_index_list += flatten([i for inst in schedule for n, i in inst.input_kwargs_dict.items()])
        # remove topmost inputs from constants
        index_list = [i for i in input_index_list if i not in trace.model_input_index_list]
        # remove values generated as output from constants
        index_list = [i for i in index_list if i not in output_index_list]
        values = {i: trace.index_map[i] for i in index_list}
        # separate constants from parameters and buffers
        consts = {i: v for i, v in values.items()
                  if not isinstance(v, Parameter) and not trace.is_buffer(v)}
        params = {i: (trace.name_from_param(v), v)
                  for i, v in values.items() if isinstance(v, Parameter)}
        buffers = {i: (trace.name_from_buffer(v), v)
                   for i, v in values.items() if trace.is_buffer(v)}
        # only scalar tensors are accepted as constants, otherwise throw an error
        consts = {i: v.item() if isinstance(v, Tensor) else v for i, v in consts.items()}
        if None in consts.keys():
            raise RuntimeError("Something went wrong")
        return consts, params, buffers

    def module_from_prefix(self, output_model, prefix):
        supermodule = output_model
        prefix_list = prefix.split('.')
        for p in prefix_list:
            if p:  # exclude empty prefix
                if p in supermodule._modules:
                    supermodule = supermodule._modules[p]
                else:  # if the module doesn't exist, create it
                    m = Module()
                    supermodule._modules[p] = m
                    supermodule = m
        return supermodule

    def _solve_inline_values(self, output_model, inst: Instruction, args, consts, buffers, params):

        def recurse(arg):
            if isinstance(arg, list):
                return [recurse(a) for a in arg]
            elif isinstance(arg, tuple):
                return tuple(recurse(a) for a in arg)
            elif isinstance(arg, dict):
                return {k: recurse(v) for k, v in arg.items()}
            elif isinstance(arg, Index):
                if arg in buffers:
                    name, buffer = buffers[arg]
                    if '.' in name:
                        final_name = name.split('.')[-1]
                        assert inst.prefix + '.' + final_name == name  # sanity check
                        name = final_name
                    module = self.module_from_prefix(output_model, inst.prefix)
                    module.register_buffer(name, buffer)
                    return buffer
                elif arg in params:
                    name, param = params[arg]
                    if '.' in name:
                        final_name = name.split('.')[-1]
                        assert inst.prefix + '.' + final_name == name  # sanity check
                        name = final_name
                    module = self.module_from_prefix(output_model, inst.prefix)
                    module.register_parameter(name, param)
                    return param
                elif arg in consts:
                    return consts[arg]
                else:
                    return arg
            else:
                return arg

        return recurse(args)

    def _apply_inline_values(self, output_model, schedule, consts, buffers, params):
        for inst in schedule:
            inst.input_args_list = self._solve_inline_values(
                output_model, inst, inst.input_args_list, consts, buffers, params)
            inst.input_kwargs_dict = self._solve_inline_values(
                output_model, inst, inst.input_kwargs_dict, consts, buffers, params)

    # Gen unused params/buffers that are not part of the schedule
    # but show up in the original state dict
    # e.g. num_batches_tracked in batch norm
    def _solve_unused_params_buffers(self, output_model, trace, used_params, used_buffers):
        used_params_names = [n for i, (n, v) in used_params.items()]
        used_buffers_names = [n for i, (n, v) in used_buffers.items()]
        for n, b in trace.named_buffers.items():
            if n not in used_buffers_names:
                prefix_list = n.split('.')
                prefix = '.'.join(prefix_list[:-1]) if len(prefix_list) > 1 else ''
                name = prefix_list[-1]
                module = self.module_from_prefix(output_model, prefix)
                module.register_buffer(name, b)
        for n, p in trace.named_params.items():
            if n not in used_params_names:
                prefix_list = n.split('.')
                prefix = '.'.join(prefix_list[:-1]) if len(prefix_list) > 1 else ''
                name = prefix_list[-1]
                module = self.module_from_prefix(output_model, prefix)
                module.register_parameter(name, p)

    def gen_model(self, trace: Trace):
        output_model = CodegenModule()
        schedule = self._gen_schedule(trace, output_model)
        consts, params, buffers = self._gen_consts_params_buffers(schedule, trace)
        self._apply_inline_values(output_model, schedule, consts, buffers, params)
        self._solve_unused_params_buffers(output_model, trace, params, buffers)
        output_model.schedule = schedule
        output_model.input_index_list = trace.model_input_index_list
        output_model.output_index_list = trace.model_output_index_list
        return output_model


# Copyright (c) 2019-     Xilinx, Inc              (Giuseppe Franco)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Implementation of various functions with a straight-through gradient estimators, dispatched to
either a native just-in-time compiled backend (when env ``BREVITAS_JIT=1``) or to an autograd
Function implemented in :obj:`~brevitas.function.autograd_ste_ops` (when env ``BREVITAS_JIT=0``).

The native backend is enabled when ``BREVITAS_JIT`` is enabled to allow for end-to-end compilation
of the built-in quantizers, since as of Pytorch 1.8.1 a torch.autograd.Function is not supported by
the compiler.
"""

import torch
from torch import Tensor

import brevitas

__all__ = [
    'round_ste',
    'ceil_ste',
    'floor_ste',
    'tensor_clamp_ste',
    'tensor_clamp_ste_',
    'scalar_clamp_ste',
    'scalar_clamp_min_ste',
    'binary_sign_ste',
    'ternary_sign_ste',
    'round_to_zero_ste',
    'abs_binary_sign_grad'
]


if brevitas.NATIVE_STE_BACKEND_LOADED:
    fn_prefix = torch.ops.autograd_ste_ops
    script_flag = brevitas.jit.script
else:
    from brevitas.function import autograd_ste_ops as fn_prefix
    script_flag = torch.jit.ignore


@script_flag
def round_ste(x: Tensor) -> Tensor:
    """
    Function that implements :func:`torch.round` with a straight-through gradient estimator.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.round_ste_impl` (with env
        ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.7, -1.7], requires_grad=True)
        >>> y = round_ste(x)
        >>> y
        tensor([ 2., -2.], grad_fn=<RoundSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.round_ste_impl(x)


@script_flag
def ceil_ste(x: Tensor) -> Tensor:
    """
    Function that implements :func:`torch.ceil` with a straight-through gradient estimator.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.ceil_ste_impl` (with env
        ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.7, -1.7], requires_grad=True)
        >>> y = ceil_ste(x)
        >>> y
        tensor([ 2., -1.], grad_fn=<CeilSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.ceil_ste_impl(x)


@script_flag
def floor_ste(x: Tensor) -> Tensor:
    """
    Function that implements :func:`torch.floor` with a straight-through gradient estimator.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.floor_ste_impl` (with env
        ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.7, -1.7], requires_grad=True)
        >>> y = floor_ste(x)
        >>> y
        tensor([ 1., -2.], grad_fn=<FloorSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.floor_ste_impl(x)


@script_flag
def tensor_clamp_ste(x: Tensor, min_val: Tensor, max_val: Tensor) -> Tensor:
    """
    Function that implements :func:`~brevitas.function.ops.tensor_clamp` with a straight-through
    gradient estimator for the gradient of y w.r.t. to x, while the gradient of y w.r.t. to min_val
    and max_val is always None.
    
    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.tensor_clamp_ste_impl` (with 
        env ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with
        ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.5, 0.4, -1.5], requires_grad=True)
        >>> y = tensor_clamp_ste(x, torch.tensor([-1.0, -0.5, -0.5]), torch.tensor([1.0, 0.5, 0.5]))
        >>> y
        tensor([ 1.0000,  0.4000, -0.5000], grad_fn=<TensorClampSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1, 0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    output = fn_prefix.tensor_clamp_ste_impl(x, min_val, max_val)
    return output


@script_flag
def tensor_clamp_ste_(x: Tensor, min_val: Tensor, max_val: Tensor) -> Tensor:
    """
    Function that implements :func:`~brevitas.function.ops.tensor_clamp_` with a straight-through
    gradient estimator for the gradient of y w.r.t. to x, while the gradient of y w.r.t. to min_val
    and max_val is always None.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.tensor_clamp_ste_impl_` (with
        env ``BREVITAS_JIT=0``) or its C++ just-in-time compiled variant (with ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.5, 0.4, -1.5], requires_grad=True)
        >>> y = tensor_clamp_ste_(x, torch.tensor([-1.0, -0.5, -0.5]), torch.tensor([1.0, 0.5, 0.5]))
        >>> y
        tensor([ 1.0000,  0.4000, -0.5000], grad_fn=<InplaceTensorClampSteFnBackward>)
        >>> (y == x).all().item()
        True
        >>> grad = torch.tensor([0.1, -0.1, 0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    output = fn_prefix.tensor_clamp_ste_impl_(x, min_val, max_val)
    return output


@script_flag
def scalar_clamp_ste(x: Tensor, min_val: float, max_val: float) -> Tensor:
    """
    Function that implements :func:`torch.clamp` with a straight-through gradient estimator
    for the gradient of the output w.r.t. to ``x``, while the gradient of ``y`` w.r.t. to ``min_val``
    and ``max_val`` is always ``None``.

    Args:
        x: input tensor to clamp.
        min_val: scalar value to use as lower bound for the input tensor.
        max_val: scalar value to use as upper bound for the input tensor.

    Returns:
        Tensor: clamped output tensor.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.scalar_clamp_ste_impl`
        (with env ``BREVITAS_JIT=0``) or its C++ just-in-time compiled variant
        (with ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.5, 0.4, -1.5], requires_grad=True)
        >>> y = scalar_clamp_ste(x, -1.0, 1.0)
        >>> y
        tensor([ 1.0000,  0.4000, -1.0000], grad_fn=<ScalarClampSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1, 0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.scalar_clamp_ste_impl(x, min_val, max_val)


@script_flag
def scalar_clamp_min_ste(x: Tensor, min_val: float) -> Tensor:
    """
    Function that implements :func:`torch.clamp_min` with a straight-through gradient estimator
    for the gradient of output y w.r.t. to ``x``, while the gradient of y w.r.t. to ``min_val`` is
    always ``None``.

    Args:
        x: input tensor to clamp.
        min_val: scalar value to use as lower bound for the input tensor.

    Returns:
        Tensor: clamped output tensor.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.scalar_clamp_min_ste_impl`
        (with env ``BREVITAS_JIT=0``) or its C++ just-in-time compiled variant
        (with ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.5, 0.4, -1.5], requires_grad=True)
        >>> y = scalar_clamp_min_ste(x, -1.0)
        >>> y
        tensor([ 1.5000,  0.4000, -1.0000], grad_fn=<ScalarClampMinSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1, 0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.scalar_clamp_min_ste_impl(x, min_val)


@script_flag
def binary_sign_ste(x: Tensor) -> Tensor:
    """
    Function that implements :func:`~brevitas.function.ops.binary_sign` with a straight-through
    gradient estimator.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.binary_sign_ste_impl` (with
        env ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with
        ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.7, 0.0, -0.5], requires_grad=True)
        >>> y = binary_sign_ste(x)
        >>> y
        tensor([ 1.,  1., -1.], grad_fn=<BinarySignSteFnBackward>)
        >>> grad = torch.tensor([0.1, 0.2, -0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.binary_sign_ste_impl(x)


@script_flag
def ternary_sign_ste(x: Tensor) -> Tensor:
    """
    Function that implements :func:`torch.sign` with a straight-through gradient estimator.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.ternary_sign_ste_impl` (with
        env ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with
        ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.7, 0.0, -0.5], requires_grad=True)
        >>> y = ternary_sign_ste(x)
        >>> y
        tensor([ 1.,  0., -1.], grad_fn=<TernarySignSteFnBackward>)
        >>> grad = torch.tensor([0.1, 0.2, -0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.ternary_sign_ste_impl(x)


@script_flag
def round_to_zero_ste(x: Tensor) -> Tensor:
    """
    Function that implements :func:`~brevitas.function.ops.round_to_zero` with a straight-through
    gradient estimator.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.round_to_zero_ste_impl` (with
        env ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with
        ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([1.7, -1.7], requires_grad=True)
        >>> y = round_to_zero_ste(x)
        >>> y
        tensor([ 1., -1.], grad_fn=<RoundToZeroSteFnBackward>)
        >>> grad = torch.tensor([0.1, -0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.round_to_zero_ste_impl(x)


@script_flag
def abs_binary_sign_grad(x: Tensor) -> Tensor:
    """
    Function that implements :func:`torch.abs` with a binary-sign backward, in order to
    have subgradient 1 in 0. Compare with :func:`torch.abs`' subgradient of 0 in 0.

    Notes:
        Wrapper for either :func:`~brevitas.function.autograd_ste_ops.abs_binary_sign_grad_impl`
        (with env ``BREVITAS_JIT=0``) or its native just-in-time compiled variant (with 
        ``BREVITAS_JIT=1``).

    Examples:
        >>> x = torch.tensor([0.0], requires_grad=True)
        >>> y = abs_binary_sign_grad(x)
        >>> y
        tensor([0.], grad_fn=<AbsBinarySignGradFnBackward>)
        >>> grad = torch.tensor([0.1])
        >>> y.backward(grad)
        >>> (x.grad == grad).all().item()
        True
    """
    return fn_prefix.abs_binary_sign_grad_impl(x)


from contextlib import ExitStack
from _collections_abc import Iterable
from inspect import getcallargs
from typing import List, Union, Any
from copy import copy
import functools
from packaging import version
from dataclasses import field, dataclass, replace

import torch
from torch import Tensor
from torch.nn import Module, Sequential, ModuleList, ModuleDict

from brevitas import torch_version
from brevitas.quant_tensor import QuantTensor
from .trace import Trace, TraceElem
from .wrapper.scriptmodule import torchscript_wrapper
from .wrapper.builtin import IntWrapper, StrWrapper, FloatWrapper
from ..module import FnType
from ..utils import flatten, module_class_name

if torch_version > version.parse('1.6'):
    from torch.overrides import get_testing_overrides
    from .patch import ABOVE_16_PATCHES as PATCHES
elif torch_version == version.parse('1.6'):
    from torch._overrides import get_testing_overrides
    from .patch import EQUAL_16_PATCHES as PATCHES
else:
    from .backport.signatures import get_testing_overrides
    from .patch import BELOW_16_PATCHES as PATCHES

TORCH_FN_NAMES = [fn.__name__ for fn in get_testing_overrides().keys()]
TORCH_FN_OVERRIDE_DICT = get_testing_overrides()

TYPES_TO_TRACE = [
    IntWrapper,
    FloatWrapper,
    StrWrapper,
    Iterable,  # works with tuple_iterator and such
    list,
    dict,
    tuple,
    torch.Tensor,
    QuantTensor,
    torch.Size]


class CallableWrapper(object):

    def __init__(self, tracer, callabl, inplace):
        self.tracer = tracer
        self.callabl = callabl
        self.inplace = inplace

    def __call__(self, *args, **kwargs):
        args, kwargs = self.tracer.repack_args_kwargs(args, kwargs)
        out = self.callabl(*args, **kwargs)
        kwargs['self'] = self.tracer.value_  # add tensor to kwargs
        out = self.tracer.update_trace(self.callabl.__name__, FnType.METHOD, args, kwargs, out)
        return self.tracer.epilogue(self.inplace, out)


# Adapted from: https://bit.ly/3hYCpvJ (stackoverflow)
class TracerMeta(type):

    # Adapted from: https://code.activestate.com/recipes/496741-object-proxying/
    magic_methods = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__', '__contains__',
        '__delitem__', '__delslice__', '__div__', '__divmod__', '__eq__', '__float__',
        '__floordiv__', '__ge__', '__getitem__', '__getslice__', '__gt__', '__hash__',
        '__hex__', '__iadd__', '__iand__', '__idiv__', '__idivmod__', '__ifloordiv__',
        '__ilshift__', '__imod__', '__imul__', '__int__', '__invert__', '__ior__', '__ipow__',
        '__irshift__', '__isub__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__', '__neg__', '__oct__',
        '__or__', '__pos__', '__pow__', '__radd__', '__rand__', '__rdiv__', '__rdivmod__',
        '__reduce__', '__reduce_ex__', '__reversed__', '__rfloorfiv__',  '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', '__rtruediv__',
        '__rxor__', '__setitem__', '__setslice__', '__sub__', '__truediv__', '__xor__',
        '__next__', '__iter__']

    @staticmethod
    def _magic_function(tracer, method_name, *args, **kwargs):
        if hasattr(tracer.value_, method_name) and method_name in TracerMeta.magic_methods:
            fn = getattr(tracer.value_, method_name)
            args, kwargs = tracer.repack_args_kwargs(args, kwargs)
            out = fn(*args, **kwargs)
            kwargs['self'] = tracer.value_
            out, inplace = tracer.update_inplace_output(out, args, kwargs)
            out = tracer.update_trace(method_name, FnType.METHOD, args, kwargs, out)
            return tracer.epilogue(inplace, out)

    def __new__(cls, name, bases, attr):
        new = super(TracerMeta, cls).__new__(cls, name, bases, attr)
        for method_name in TracerMeta.magic_methods:
            magic_method = functools.partialmethod(TracerMeta._magic_function, method_name)
            setattr(new, method_name, magic_method)
        return new


class Tracer(metaclass=TracerMeta):

    def __init__(self, value_, trace_ = None, namespace_= None):
        self.value_ = value_
        self.trace_ = trace_ or Trace()
        self.namespace_ = namespace_ or []
        self.hook_handles = []
        self.preserve_module_allowlist: List = ['torch.nn', 'brevitas.nn']
        self.preserve_module_blocklist: List = [Sequential, ModuleList, ModuleDict]

    def __bool__(self):
        return self.value_

    def __getattr__(self, name: str):
        inplace_tensor = name.endswith('_') and name[:-1] in TORCH_FN_NAMES
        attr = getattr(self.value_, name)
        if callable(attr):
            return CallableWrapper(self, attr, inplace_tensor)
        else:
            kwargs = {'self': self.value_}
            self.update_trace(name, FnType.ATTRIBUTE, [], kwargs, attr)
            return self.epilogue(False, attr)

    def trace_module_through(self, module):
        return (isinstance(module, tuple(self.preserve_module_blocklist))
                or all([not module_class_name(module).startswith(pma)
                        for pma in self.preserve_module_allowlist]))

    def is_tracing(self, input_list):
        return any(isinstance(i, Tracer) for i in flatten(input_list))

    def repack_value(self, value):
        if self.trace_.index_from_map(value) is None:
            if isinstance(value, Tracer):
                return value.value_
            elif type(value) is int:
                return IntWrapper(value)
            elif type(value) is float:
                return FloatWrapper(value)
            elif type(value) is str:
                return StrWrapper(value)
            elif isinstance(value, tuple) and not isinstance(value, QuantTensor):
                return tuple(self.repack_value(v) for v in value)
            elif isinstance(value, list):
                return [self.repack_value(v) for v in value]
            elif isinstance(value, dict):
                return {self.repack_value(k): self.repack_value(v) for k, v in value.items()}
            else:
                return value
        else:
            return value

    def repack_args_kwargs(self, args, kwargs):
        new_args = [self.repack_value(a) for a in args]
        new_kwargs = {k: self.repack_value(a) for k, a in kwargs.items()}
        return new_args, new_kwargs

    def repack_module_input(self, value):
        if isinstance(value, list) or (isinstance(value, tuple) and len(value) == 1):
            value = [self.repack_value(v) for v in value]
        else:
            value = [self.repack_value(value)]
        return value

    def repack_model_output(self, value):
        if isinstance(value, (list, tuple)):
            value = self.repack_value(value)
        else:
            value = [self.repack_value(value)]
        return value

    def epilogue(self, inplace, output):
        if inplace:
            assert not isinstance(output, tuple)  # sanity check
            self.value_ = output
            return self
        elif isinstance(output, tuple(TYPES_TO_TRACE)):
            new_tracer = Tracer(trace_=self.trace_, namespace_=self.namespace_, value_=output)
            return new_tracer
        else:
            return output

    def _register_forward_pre_hooks(self, model: torch.nn.Module):

        def enter_namespace(module, input):
            name = None
            if not self.namespace_:  # root module
                name = ''
            else:
                supermodule = self.namespace_[-1][0]
                for name, mod in supermodule.named_modules():
                    # identify current module as a submodule
                    if mod is module:
                        break
            input_list = self.repack_module_input(input)
            namespace = (module, name, input_list, input)
            self.namespace_.append(namespace)
            if not self.trace_module_through(module):
                return tuple(input_list)

        # Identify input to the whole model
        def model_input(module, input):
            if module is model:
                input_list = self.repack_module_input(input)
                self.trace_.model_input_list = input_list

        for mod in model.modules():
            if not isinstance(mod, torch.jit.ScriptModule):
                self.hook_handles.append(mod.register_forward_pre_hook(enter_namespace))
                self.hook_handles.append(mod.register_forward_pre_hook(model_input))

    def _register_forward_hooks(self, model):

        # input is possibly the Tracer
        def exit_namespace(module, input, output):
            assert self.namespace_[-1][0] is module  # check module is on the stack
            for trace_elem in self.trace_.trace_elem_list:
                if trace_elem.module_context_list[-1] is module:
                    if trace_elem.module_output is None:
                        output = self.repack_value(output)
                        trace_elem.module_output = output
            namespace = self.namespace_.pop()
            orig_input = namespace[3]
            # is_tracins is to check that I'm not in an allowed module inside a blocked module
            if self.is_tracing(orig_input) and not self.trace_module_through(module):
                args = namespace[2]
                module_name = namespace[1]
                output, inplace = self.update_inplace_output(output, args, {})
                output = self.update_trace(module, FnType.MODULE, tuple(args), {}, output)
                self.trace_.trace_elem_list[-1].module_fn_name = module_name
                if inplace:
                    assert isinstance(orig_input, tuple) and len(orig_input) == 1
                    input_tracer = orig_input[0]
                    # self is the tracer from when the hook was registered, not when the input
                    # is passed. it doesn't make a difference if the operation is not in place,
                    # but if the operation is in place i need to update the value_ of the tracer
                    # that was passed as input
                    return input_tracer.epilogue(True, output)
                else:
                    return self.epilogue(False, output)

        # Identify output of the whole model
        def model_output(module, input, output):
            if module is model:
                self.trace_.model_output_list = self.repack_model_output(output)
                return self.repack_value(output)

        for mod in model.modules():
            if not isinstance(mod, torch.jit.ScriptModule):
                self.hook_handles.append(mod.register_forward_hook(exit_namespace))
                self.hook_handles.append(mod.register_forward_hook(model_output))

    def move_torch_args_to_kwargs(self, fn, fn_args, fn_kwargs):
        fn_stub = TORCH_FN_OVERRIDE_DICT[fn]
        fn_kwargs = getcallargs(fn_stub, *fn_args, **fn_kwargs)
        return [], fn_kwargs

    def update_inplace_output(self, out, fn_args, fn_kwargs):
        inplace = any([out is arg for arg in fn_args + list(fn_kwargs.values())])
        out = out.clone() if inplace and isinstance(out, Tensor) else out
        return out, inplace

    def __torch_function__(self, fn, types, args=(), kwargs=None):
        kwargs = {} if kwargs is None else kwargs
        args, kwargs = self.move_torch_args_to_kwargs(fn, args, kwargs)
        args, kwargs = self.repack_args_kwargs(args, kwargs)
        # get rid of out=None early on since it's often problematic
        if 'out' in kwargs and kwargs['out'] is None: del kwargs['out']
        out = fn(*args, **kwargs)
        out, inplace = self.update_inplace_output(out, args, kwargs)
        out = self.update_trace(fn, FnType.FUNCTION, args, kwargs, out)
        return self.epilogue(inplace, out)

    def update_trace(self, fn, fn_type, fn_args, fn_kwargs, fn_out):
        modules, m_names, m_inputs, _ = zip(*self.namespace_)
        m_names = flatten(m_names)
        # remove first empty name
        m_names = m_names[1:]
        # take only innermost input list, corresponding to modules[-1]
        m_input_list = m_inputs[-1]
        # empty fn_args means it would be an empty tuple, which we don't want
        fn_args = [] if fn_args == () else fn_args
        # repack fn_out
        fn_out = self.repack_value(fn_out)
        # assign index to fn inputs and outputs
        fn_args_index = [self.trace_.index_from_val(a) for a in fn_args]
        fn_kwargs_index = {k: self.trace_.index_from_val(a) for k,a in fn_kwargs.items()}
        fn_out_index = self.trace_.index_from_val(fn_out, False)
        # generate trace element
        trace_elem = TraceElem(
            fn=fn,
            fn_type=fn_type,
            fn_args=fn_args,
            fn_kwargs=fn_kwargs,
            fn_out=fn_out,
            fn_args_index=fn_args_index,
            fn_kwargs_index=fn_kwargs_index,
            fn_out_index=fn_out_index,
            module_context_list=modules,
            prefix_list=m_names,
            module_input_list=m_input_list)
        # update trace
        self.trace_.trace_elem_list.append(trace_elem)
        # return fn_out with newly interned values
        return fn_out

    # https://stackoverflow.com/questions/3024925/create-a-with-block-on-several-context-managers
    def _trace_with_patches(self, model):
        with ExitStack() as stack:
            for mgr in PATCHES:
                stack.enter_context(mgr)
            model(self)

    def trace_model(self, model: Module, wrap_torchscript=True):
        model = copy(model)
        if wrap_torchscript:
            torchscript_wrapper(model)
        self._register_forward_pre_hooks(model)
        self._register_forward_hooks(model)
        self._trace_with_patches(model)
        self.trace_.named_buffers = dict(model.named_buffers())
        self.trace_.named_params = dict(model.named_parameters())
        for h in self.hook_handles:
            h.remove()
        return self.trace_



# Adapted from https://github.com/NVIDIA/NeMo/blob/r0.9/collections/nemo_asr/
# Copyright (C) 2020 Xilinx (Giuseppe Franco)
# Copyright (C) 2019 NVIDIA CORPORATION.
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch
import torch.nn as nn


class CTCLossNM(nn.Module):
    """
    Neural Module wrapper for pytorch's ctcloss

    Args:
        num_classes (int): Number of characters in ASR model's vocab/labels.
            This count should not include the CTC blank symbol.
    """

    def __init__(self, *, num_classes, **kwargs):
        nn.Module.__init__(self)

        # self._blank = self.local_parameters.get('blank', 0)
        self._blank = num_classes
        self._criterion = nn.CTCLoss(blank=self._blank,
                                     reduction='none')

    def _loss(self, log_probs, targets, input_length, target_length):
        input_length = input_length.long()
        target_length = target_length.long()
        targets = targets.long()
        loss = self._criterion(log_probs.transpose(1, 0), targets,
                               input_length,
                               target_length)
        # note that this is different from reduction = 'mean'
        # because we are not dividing by target lengths
        loss = torch.mean(loss)
        return loss

    def _loss_function(self, **kwargs):
        return self._loss(*(kwargs.values()))



# Copyright (c) 2018-     Xilinx, Inc              (Alessandro Pappalardo)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
A collection of miscellaneous ScriptModule used in various quantizers.
"""

import brevitas
import torch


class InplaceNoOp(brevitas.jit.ScriptModule):
    """
    ScriptModule in-place no-op placehoder.

    Examples:
        >>> x = torch.tensor(1.0)
        >>> no_op = InplaceNoOp()
        >>> no_op(x)
    """

    def __init__(self):
        super(InplaceNoOp, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor) -> None:
        return None


class Identity(brevitas.jit.ScriptModule):
    """
    Identity ScriptModule.

    Examples:
        >>> identity = Identity()
        >>> x = torch.randn(size=[10,])
        >>> y = identity(x)
        >>> y is x
        True
    """

    def __init__(self) -> None:
        super(Identity, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x


class PowerOfTwo(brevitas.jit.ScriptModule):
    """
    ScriptModule implementation of 2.0 ** x.

    Examples:
        >>> power_of_two = PowerOfTwo()
        >>> x = torch.tensor(5.0)
        >>> power_of_two(x)
        tensor(32.)
    """

    def __init__(self) -> None:
        super(PowerOfTwo, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return 2.0 ** x


class LogTwo(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~torch.log2`.

    Examples:
        >>> log_two = LogTwo()
        >>> x = torch.tensor(8.0)
        >>> log_two(x)
        tensor(3.)
    """

    def __init__(self) -> None:
        super(LogTwo, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.log2(x)


class InplaceLogTwo(brevitas.jit.ScriptModule):
    """
    ScriptModule wrapper for :func:`~torch.log2_`.

    Examples:
        >>> inplace_log_two = InplaceLogTwo()
        >>> x = torch.tensor(8.0)
        >>> inplace_log_two(x)
        >>> x
        tensor(3.)

    Note:
        The forward method returns None.
    """

    def __init__(self) -> None:
        super(InplaceLogTwo, self).__init__()

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor) -> None:
        x.log2_()









# Copyright (c) 2019-     Xilinx, Inc              (Giuseppe Franco)
# Copyright (c) 2018-     Xilinx, Inc              (Alessandro Pappalardo)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Union, Tuple, Type, Optional

import torch
from torch import Tensor
from torch.nn import ConvTranspose1d, ConvTranspose2d
from torch.nn.functional import conv_transpose1d, conv_transpose2d

from brevitas.function.ops import max_int
from brevitas.function.ops_ste import ceil_ste
from brevitas.quant_tensor import QuantTensor
from brevitas.inject.defaults import Int8WeightPerTensorFloat
from .quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from .quant_layer import WeightQuantType, BiasQuantType, ActQuantType


__all__ = ['QuantConvTranspose1d', 'QuantConvTranspose2d']


class QuantConvTranspose1d(QuantWBIOL, ConvTranspose1d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int,
            stride: int = 1,
            padding: int = 0,
            output_padding: int = 0,
            dilation: int = 1,
            groups: int = 1,
            bias: bool = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs) -> None:
        ConvTranspose1d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            bias=bias)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self._output_size = None

    @property
    def per_elem_ops(self):
        raise NotImplementedError

    @property
    def output_channel_dim(self) -> int:
        return 0

    @property
    def channelwise_separable(self) -> bool:
        raise self.groups == self.weight.shape[0]

    def forward(self, input: Union[Tensor, QuantTensor], output_size=None) -> Union[Tensor, QuantTensor]:
        self._output_size = output_size  # cache the value temporarily
        return self.forward_impl(input)

    def compute_output_padding(self, inp, output_size):
        return self._output_padding(
            inp, output_size, self.stride, self.padding, self.kernel_size)

    def conv_transpose1d_zeros_pad(
            self, x: Tensor, weight: Tensor, bias: Optional[Tensor], output_padding):
        out = conv_transpose1d(
            x, weight, bias, self.stride, self.padding, output_padding, self.groups, self.dilation)
        return out

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.padding_mode == 'zeros':
            output_padding = self.compute_output_padding(x, self._output_size)
            self._output_size = None  # set it back to None after consuming it
            out = self.conv_transpose1d_zeros_pad(x, quant_weight, quant_bias, output_padding)
            return out
        else:
            raise NotImplementedError(f"Padding mode {self.padding_mode} not supported.")

    def max_acc_bit_width(self, input_bit_width, weight_bit_width):
        max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
        max_kernel_val = self.weight_quant.max_uint_value(weight_bit_width)
        group_size = self.out_channels // self.groups
        overlapping_sums = max(round(self.kernel_size[0] / self.stride[0]), 1)
        max_uint_output = max_uint_input * max_kernel_val * overlapping_sums * group_size
        max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
        return max_output_bit_width


class QuantConvTranspose2d(QuantWBIOL, ConvTranspose2d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple[int]],
            stride: Union[int, Tuple[int]] = 1,
            padding: Union[int, Tuple[int]] = 0,
            output_padding: Union[int, Tuple[int]] = 0,
            dilation: Union[int, Tuple[int]] = 1,
            groups: int = 1,
            bias: bool = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            **kwargs) -> None:
        ConvTranspose2d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            bias=bias)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self._output_size = None

    @property
    def per_elem_ops(self):
        raise NotImplementedError

    @property
    def output_channel_dim(self) -> int:
        return 0

    @property
    def channelwise_separable(self) -> bool:
        raise self.groups == self.weight.shape[0]

    def forward(self, input: Union[Tensor, QuantTensor], output_size=None) -> Union[Tensor, QuantTensor]:
        self._output_size = output_size  # cache the value temporarily
        return self.forward_impl(input)

    def compute_output_padding(self, inp, output_size):
        return self._output_padding(
            inp, output_size, self.stride, self.padding, self.kernel_size)

    def conv_transpose2d_zeros_pad(
            self, x: Tensor, weight: Tensor, bias: Optional[Tensor], output_padding):
        out = conv_transpose2d(
            x, weight, bias, self.stride, self.padding, output_padding, self.groups, self.dilation)
        return out

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.padding_mode == 'zeros':
            output_padding = self.compute_output_padding(x, self._output_size)
            self._output_size = None  # set it back to None after consuming it
            out = self.conv_transpose2d_zeros_pad(x, quant_weight, quant_bias, output_padding)
            return out
        else:
            raise NotImplementedError(f"Padding mode {self.padding_mode} not supported.")

    def max_acc_bit_width(self, input_bit_width, weight_bit_width):
        max_uint_input = max_int(bit_width=input_bit_width, signed=False, narrow_range=False)
        max_kernel_val = self.weight_quant.max_uint_value(weight_bit_width)
        group_size = self.out_channels // self.groups
        overlapping_sums = max(round(self.kernel_size[0] / self.stride[0]), 1)
        overlapping_sums *= max(round(self.kernel_size[1] / self.stride[1]), 1)
        max_uint_output = max_uint_input * max_kernel_val * overlapping_sums * group_size
        max_output_bit_width = ceil_ste(torch.log2(max_uint_output))
        return max_output_bit_width



import os
from configparser import ConfigParser

from torch import hub

from .melgan import *

model_impl = {
    'melgan': melgan,
}


def model_with_cfg(name, pretrained):
    cfg = ConfigParser()
    current_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(current_dir, '..', 'cfg', name + '.ini')
    assert os.path.exists(config_path)
    cfg.read(config_path)
    arch = cfg.get('MODEL', 'ARCH')
    model = model_impl[arch](cfg)
    if pretrained:
        checkpoint = cfg.get('MODEL', 'PRETRAINED_URL')
        state_dict = hub.load_state_dict_from_url(checkpoint, progress=True, map_location='cpu')
        model.load_state_dict(state_dict, strict=True)
    return model, cfg


def quant_melgan_8b(pretrained=True):
    model, _ = model_with_cfg('quant_melgan_8b', pretrained)
    return model



# Copyright (c) 2018-     Xilinx, Inc              (Alessandro Pappalardo)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from torch.nn import Sequential


class TupleSequential(Sequential):

    def output(self, mod, input):
        if isinstance(input, tuple):
            return mod(*input)
        else:
            return mod(input)

    def forward(self, *input):
        modules = list(self._modules.values())
        out = self.output(modules[0], input)
        for mod in modules[1:]:
            out = self.output(mod, out)
        return out


# Adapted from https://github.com/NVIDIA/NeMo/blob/r0.9/collections/nemo_asr/
# Copyright (C) 2020 Xilinx (Giuseppe Franco)
# Copyright (C) 2019 NVIDIA CORPORATION.
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import torch
import torch.nn as nn


class SpecAugment(nn.Module):
    """
    Zeroes out(cuts) random continuous horisontal or
    vertical segments of the spectrogram as described in
    SpecAugment (https://arxiv.org/abs/1904.08779).

    params:
    freq_masks - how many frequency segments should be cut
    time_masks - how many time segments should be cut
    freq_width - maximum number of frequencies to be cut in one segment
    time_width - maximum number of time steps to be cut in one segment
    """
    def __init__(
        self,
        freq_masks=0,
        time_masks=0,
        freq_width=10,
        time_width=10,
        rng=None
    ):
        super(SpecAugment, self).__init__()

        self._rng = random.Random() if rng is None else rng

        self.freq_masks = freq_masks
        self.time_masks = time_masks

        self.freq_width = freq_width
        self.time_width = time_width

    @torch.no_grad()
    def forward(self, x):
        sh = x.shape

        mask = torch.zeros(x.shape).byte()

        for idx in range(sh[0]):
            for i in range(self.freq_masks):
                x_left = int(self._rng.uniform(
                    0, sh[1] - self.freq_width))

                w = int(self._rng.uniform(0, self.freq_width))

                mask[idx, x_left:x_left + w, :] = 1

            for i in range(self.time_masks):
                y_left = int(self._rng.uniform(
                    0, sh[2] - self.time_width))

                w = int(self._rng.uniform(0, self.time_width))

                mask[idx, :,
                     y_left:y_left + w] = 1

        x = x.masked_fill(mask.type(torch.bool).to(device=x.device), 0)

        return x


class SpecCutout(nn.Module):
    """
    Zeroes out(cuts) random rectangles in the spectrogram
    as described in (https://arxiv.org/abs/1708.04552).

    params:
    rect_masks - how many rectangular masks should be cut
    rect_freq - maximum size of cut rectangles along the frequency dimension
    rect_time - maximum size of cut rectangles along the time dimension
    """
    def __init__(
        self,
        rect_masks=0,
        rect_time=5,
        rect_freq=20,
        rng=None
    ):
        super(SpecCutout, self).__init__()

        self._rng = random.Random() if rng is None else rng

        self.rect_masks = rect_masks
        self.rect_time = rect_time
        self.rect_freq = rect_freq

    @torch.no_grad()
    def forward(self, x):
        sh = x.shape

        mask = torch.zeros(x.shape).byte()

        for idx in range(sh[0]):
            for i in range(self.rect_masks):
                rect_x = int(self._rng.uniform(
                    0, sh[1] - self.rect_freq))
                rect_y = int(self._rng.uniform(
                    0, sh[2] - self.rect_time))

                w_x = int(self._rng.uniform(0, self.rect_time))
                w_y = int(self._rng.uniform(0, self.rect_freq))

                mask[idx, rect_x:rect_x + w_x,
                     rect_y:rect_y + w_y] = 1

        x = x.masked_fill(mask.type(torch.bool).to(device=x.device), 0)

        return x



from typing import ClassVar, Optional, List
from inspect import isclass
from enum import auto

import torch
from torch import nn

from brevitas import nn as qnn
from brevitas.inject.defaults import *
from .tracer import Tracer
from .generator import ModuleGenerator
from .rewriter import TorchFnToModuleRewriter, TensorMethodToModuleRewriter
from .rewriter import ModuleToModuleRewriter
from .rewriter import MergeBatchNorm2d, DisableBreakingReturnQuantTensor
from .rewriter import DuplicateSharedStatelessModule
from .rewriter import MeanToAdaptiveAvgPool2d, InplaceMeanToAdaptiveAvgPool2d
from brevitas.utils.python_utils import AutoName


_io_map = {
    '__add__': qnn.QuantEltwiseAdd,  # todo this should be restricted to tensor types only
    '__iadd__': qnn.QuantEltwiseAdd,
    'add': qnn.QuantEltwiseAdd,
    'add_': qnn.QuantEltwiseAdd,
    torch.add: qnn.QuantEltwiseAdd,
    torch.cat: qnn.QuantCat}


_wbiol_map = {
    nn.Conv1d: qnn.QuantConv1d,
    nn.Conv2d: qnn.QuantConv2d,
    nn.ConvTranspose1d: qnn.QuantConvTranspose1d,
    nn.ConvTranspose2d: qnn.QuantConvTranspose2d,
    nn.Linear: qnn.QuantLinear}


_act_map = {
    nn.ReLU: qnn.QuantReLU,
    nn.ReLU6: qnn.QuantReLU,
    nn.Hardtanh: qnn.QuantHardTanh,
    nn.Sigmoid: qnn.QuantSigmoid,
    nn.Tanh: qnn.QuantTanh,
    torch.relu: qnn.QuantReLU,
    torch.relu_: qnn.QuantReLU,
    torch.nn.functional.relu: qnn.QuantReLU,
    torch.nn.functional.relu_: qnn.QuantReLU,
    torch.nn.functional.relu6: qnn.QuantReLU,
    'relu': qnn.QuantReLU,
    'relu_': qnn.QuantReLU}  # todo add more functional apis


_trunc_map = {
    nn.AvgPool2d: qnn.QuantAvgPool2d,
    nn.AdaptiveAvgPool2d: qnn.QuantAdaptiveAvgPool2d,
    torch.nn.functional.avg_pool2d: qnn.QuantAvgPool2d,
    torch.nn.functional.adaptive_avg_pool2d: qnn.QuantAdaptiveAvgPool2d}


_wrapper_map = {
    nn.MaxPool1d: qnn.QuantMaxPool1d,
    nn.MaxPool2d: qnn.QuantMaxPool2d,
    nn.Dropout: qnn.QuantDropout}


_upsample_map = {
    nn.Upsample: qnn.QuantUpsample,
    nn.UpsamplingBilinear2d: qnn.QuantUpsamplingBilinear2d,
    nn.UpsamplingNearest2d: qnn.QuantUpsamplingNearest2d}


def _map_to_rewriter_impl(original):
    if isinstance(original, str):
        return TensorMethodToModuleRewriter
    elif isclass(original) and issubclass(original, nn.Module):
        return ModuleToModuleRewriter
    elif callable(original):
        return TorchFnToModuleRewriter
    else:
        raise RuntimeError(f"Original {original} type not recognized")


class BatchNormHandling(AutoName):
    PRESERVE = auto()
    QUANTIZE = auto()
    MERGE_AND_QUANTIZE = auto()
    MERGE_AND_PRESERVE = auto()


def wbiol_rewriter_list(
        input_quant, weight_quant, bias_quant, output_quant, return_quant_tensor, excluded, **kwargs):
    iq, wq, bq, oq = input_quant, weight_quant, bias_quant, output_quant
    rewriter_list = [_map_to_rewriter_impl(original)(
        original, quant, input_quant=iq, weight_quant=wq, bias_quant=bq, output_quant=oq,
        return_quant_tensor=return_quant_tensor, **kwargs)
        for original, quant in _wbiol_map.items() if original not in excluded]
    return rewriter_list


def act_rewriter_list(act_quant, return_quant_tensor, excluded, **kwargs):
    aq = act_quant
    rewriter_list = [_map_to_rewriter_impl(original)(
        original, quant, act_quant=aq, return_quant_tensor=return_quant_tensor, **kwargs)
        for original, quant in _act_map.items() if original not in excluded]
    return rewriter_list


def trunc_rewriter_list(trunc_quant, return_quant_tensor, excluded, **kwargs):
    rewriter_list = [_map_to_rewriter_impl(original)(
        original, quant, trunc_quant=trunc_quant,
        return_quant_tensor=return_quant_tensor, **kwargs)
        for original, quant in _trunc_map.items() if original not in excluded]
    if 'mean' not in excluded:
        rewriter_list.append(MeanToAdaptiveAvgPool2d())
    if 'mean_' not in excluded:
        rewriter_list.append(InplaceMeanToAdaptiveAvgPool2d())
    return rewriter_list


def no_args_rewriter_list(return_quant_tensor, excluded):
    wrapper_list = [_map_to_rewriter_impl(original)(
        original, quant, return_quant_tensor=return_quant_tensor)
        for original, quant in _wrapper_map.items() if original not in excluded]
    upsample_list = [_map_to_rewriter_impl(original)(
        original, quant, return_quant_tensor=return_quant_tensor)
        for original, quant in _upsample_map.items() if original not in excluded]
    return wrapper_list + upsample_list


def io_rewriter_list(input_quant, output_quant, return_quant_tensor, excluded, **kwargs):
    rewriter_list = [_map_to_rewriter_impl(original)(
        original, quant, input_quant=input_quant, output_quant=output_quant,
        return_quant_tensor=return_quant_tensor, **kwargs)
        for original, quant in _io_map.items() if original not in excluded]
    return rewriter_list


def quantize(
        model: nn.Module,
        input: torch.Tensor,
        weight_quant = Int8WeightPerTensorFloat,
        input_quant = Int8ActPerTensorFloat,
        output_quant = Int8ActPerTensorFloat,
        trunc_quant = TruncTo8bit,
        bias_quant = Int8Bias,
        act_quant = None,
        bn_handling: BatchNormHandling = BatchNormHandling.MERGE_AND_QUANTIZE,
        excluded: Optional[List] = (),
        **kwargs):
    iq, oq, bq = input_quant, output_quant, bias_quant
    wq, aq, tq = weight_quant, act_quant, trunc_quant
    with torch.no_grad():
        trace = Tracer(input).trace_model(model)
        gen_model = ModuleGenerator().gen_model(trace)
    state_dict = model.state_dict()
    gen_model = DuplicateSharedStatelessModule().apply(gen_model)
    for rewriter in wbiol_rewriter_list(iq, wq, bq, oq, True, excluded, **kwargs):
        gen_model = rewriter.apply(gen_model)
    for rewriter in act_rewriter_list(aq, True, excluded, **kwargs):
        gen_model = rewriter.apply(gen_model)
    for rewriter in trunc_rewriter_list(tq, True, excluded, **kwargs):
        gen_model = rewriter.apply(gen_model)
    for rewriter in io_rewriter_list(iq, oq, True, excluded, **kwargs):
        gen_model = rewriter.apply(gen_model)
    for rewriter in no_args_rewriter_list(True, excluded):
        gen_model = rewriter.apply(gen_model)
    gen_model.load_state_dict(state_dict)
    if (bn_handling == BatchNormHandling.MERGE_AND_QUANTIZE
            or bn_handling == BatchNormHandling.MERGE_AND_PRESERVE):
        gen_model = MergeBatchNorm2d().apply(gen_model)
    if (bn_handling == BatchNormHandling.QUANTIZE
            or bn_handling == BatchNormHandling.MERGE_AND_QUANTIZE):
        rewriter = ModuleToModuleRewriter(
            nn.BatchNorm2d,
            qnn.BatchNorm2dToQuantScaleBias,
            input_quant=input_quant,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            output_quant=output_quant,
            return_quant_tensor=True,
            **kwargs)
        gen_model = rewriter.apply(gen_model)
    gen_model = DisableBreakingReturnQuantTensor().apply(gen_model)
    return gen_model




from .const import BitWidthConst, MsbClampBitWidth
from .parameter import BitWidthParameter, RemoveBitwidthParameter

# retrocompatibility
# the assert prevents the removal of the unused import
from brevitas.inject.enum import BitWidthImplType
assert BitWidthImplType


# Adapted from https://github.com/NVIDIA/NeMo/tree/r0.9
# Copyright (C) 2020 Xilinx (Giuseppe Franco)
# Copyright (C) 2019 NVIDIA CORPORATION.
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import fnmatch
import json
import os
import subprocess
import tarfile
import urllib.request

from sox import Transformer
from tqdm import tqdm

parser = argparse.ArgumentParser(description='LibriSpeech Data download')
parser.add_argument("--data_root", required=True, default=None, type=str)
parser.add_argument("--data_sets", default="dev_clean", type=str)
args = parser.parse_args()

URLS = {
    'TRAIN_CLEAN_100': ("http://www.openslr.org/resources/12/train-clean-100.tar.gz"),
    'TRAIN_CLEAN_360': ("http://www.openslr.org/resources/12/train-clean-360.tar.gz"),
    'TRAIN_OTHER_500': ("http://www.openslr.org/resources/12/train-other-500.tar.gz"),
    'DEV_CLEAN': "http://www.openslr.org/resources/12/dev-clean.tar.gz",
    'DEV_OTHER': "http://www.openslr.org/resources/12/dev-other.tar.gz",
    'TEST_CLEAN': "http://www.openslr.org/resources/12/test-clean.tar.gz",
    'TEST_OTHER': "http://www.openslr.org/resources/12/test-other.tar.gz",
}


def __maybe_download_file(destination: str, source: str):
    """
    Downloads source to destination if it doesn't exist.
    If exists, skips download
    Args:
        destination: local filepath
        source: url of resource

    Returns:

    """
    source = URLS[source]
    if not os.path.exists(destination):
        print("{0} does not exist. Downloading ...".format(destination))
        urllib.request.urlretrieve(source, filename=destination + '.tmp')
        os.rename(destination + '.tmp', destination)
        print("Downloaded {0}.".format(destination))
    else:
        print("Destination {0} exists. Skipping.".format(destination))
    return destination


def __extract_file(filepath: str, data_dir: str):
    try:
        tar = tarfile.open(filepath)
        tar.extractall(data_dir)
        tar.close()
    except Exception:
        print('Not extracting. Maybe already there?')


def __process_data(data_folder: str, dst_folder: str, manifest_file: str):
    """
    Converts flac to wav and build manifests's json
    Args:
        data_folder: source with flac files
        dst_folder: where wav files will be stored
        manifest_file: where to store manifest

    Returns:

    """

    if not os.path.exists(dst_folder):
        os.makedirs(dst_folder)

    files = []
    entries = []

    for root, dirnames, filenames in os.walk(data_folder):
        for filename in fnmatch.filter(filenames, '*.trans.txt'):
            files.append((os.path.join(root, filename), root))

    for transcripts_file, root in tqdm(files):
        with open(transcripts_file, encoding="utf-8") as fin:
            for line in fin:
                id, text = line[: line.index(" ")], line[line.index(" ") + 1 :]
                transcript_text = text.lower().strip()

                # Convert FLAC file to WAV
                flac_file = os.path.join(root, id + ".flac")
                wav_file = os.path.join(dst_folder, id + ".wav")
                if not os.path.exists(wav_file):
                    Transformer().build(flac_file, wav_file)
                # check duration
                duration = subprocess.check_output("soxi -D {0}".format(wav_file), shell=True)

                entry = dict()
                entry['audio_filepath'] = os.path.abspath(wav_file)
                entry['duration'] = float(duration)
                entry['text'] = transcript_text
                entries.append(entry)

    with open(manifest_file, 'w') as fout:
        for m in entries:
            fout.write(json.dumps(m) + '\n')


def main():
    data_root = args.data_root
    data_sets = args.data_sets

    if data_sets == "ALL":
        data_sets = "dev_clean,dev_other,train_clean_100,train_clean_360,train_other_500,test_clean,test_other"

    for data_set in data_sets.split(','):
        print("\n\nWorking on: {0}".format(data_set))
        filepath = os.path.join(data_root, data_set + ".tar.gz")
        print("Getting {0}".format(data_set))
        __maybe_download_file(filepath, data_set.upper())
        print("Extracting {0}".format(data_set))
        __extract_file(filepath, data_root)
        print("Processing {0}".format(data_set))
        __process_data(
            os.path.join(os.path.join(data_root, "LibriSpeech"), data_set.replace("_", "-"),),
            os.path.join(os.path.join(data_root, "LibriSpeech"), data_set.replace("_", "-"),) + "-processed",
            os.path.join(data_root, data_set + ".json"),
        )
    print('Done!')


if __name__ == "__main__":
    main()


from dataclasses import replace

import torch
from torch.nn import Module

from brevitas.graph.module import FnType, _replace_module


def torchscript_wrapper(model):
    script_modules = set([])
    names, _ = zip(*list(model.named_modules()))
    prefix_list_list = [name.split(".") for name in names]
    for prefix_list in prefix_list_list:
        supermodule = model
        for prefix in prefix_list:
            if prefix != '':
                submodule = supermodule._modules[prefix]
                if not isinstance(supermodule, torch.jit.ScriptModule) \
                        and isinstance(submodule, torch.jit.ScriptModule):
                    script_modules.add(submodule)
                    break
                else:
                    supermodule = submodule
    for sm in script_modules:
        _replace_module(model, sm, ScriptModuleWrapper(sm))


class ScriptModuleWrapper(Module):

    def __init__(self, script_module):
        super().__init__()
        self.script_module = script_module

    def forward(self, x):  # TODO multiple inputs
        assert not isinstance(x, tuple), 'Multiple inputs not supported'
        if isinstance(x, Tracer):
            out = self.script_module(x.tensor_)
            if isinstance(out, tuple):
                out = [replace(x, tensor_=o) for o in out]  # they all have to be tensors
                x._update_trace(self.script_module, FnType.SCRIPTMODULE, [x.tensor_], {}, out)
            return out[0]  # TODO multiple outputs
        else:
            return self.script_module(x)


# BSD 3-Clause License
#
# Copyright (c) 2020 Xilinx, Inc (Giuseppe Franco)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import brevitas.nn as quant_nn
from brevitas.core.quant import QuantType
from brevitas.core.scaling import ScalingImplType
from brevitas.core.restrict_val import RestrictValueType
from brevitas.core.stats import StatsOp
global ACT_MIN_VAL, ACT_MAX_VAL
brevitas_activations = {
    "hardtanh": quant_nn.QuantHardTanh,
    "relu": quant_nn.QuantReLU,
}

QUANT_TYPE = QuantType.INT
QUANT_TYPE_BIAS = QuantType.FP

SCALING_MIN_VAL = 2e-16
ACT_SCALING_IMPL_TYPE = ScalingImplType.PARAMETER
ACT_SCALING_PER_CHANNEL = False
ACT_RESTRICT_SCALING_TYPE = RestrictValueType.LOG_FP

WEIGHT_SCALING_IMPL_TYPE = ScalingImplType.STATS
WEIGHT_SCALING_STATS_OP = StatsOp.MAX
WEIGHT_NARROW_RANGE = True
BIAS_CONFIGS = False


def make_quantization_input(bit_width, absolute_act_val, scaling_per_channel):
    return quant_nn.QuantHardTanh(bit_width=bit_width, scaling_per_channel=scaling_per_channel, quant_type=QUANT_TYPE,
                                  scaling_impl_type=ACT_SCALING_IMPL_TYPE, scaling_min_val=SCALING_MIN_VAL, restrict_scaling_type=ACT_RESTRICT_SCALING_TYPE,
                                  max_val=absolute_act_val, min_val=-absolute_act_val, return_quant_tensor=False)


def make_norm_scale(bit_width, absolute_act_val, scaling_per_channel):
    return quant_nn.QuantHardTanh(bit_width=bit_width, scaling_per_channel=scaling_per_channel, quant_type=QUANT_TYPE,
                                  scaling_impl_type=ACT_SCALING_IMPL_TYPE, scaling_min_val=SCALING_MIN_VAL, restrict_scaling_type=ACT_RESTRICT_SCALING_TYPE,
                                  max_val=absolute_act_val, min_val=-absolute_act_val, scaling_stats_permute_dims=(1, 0, 2),
                                  return_quant_tensor=True)


def make_jasper_activation(activation, channels, bit_width, absolute_act_val, scaling_per_channel):
    brevitas_activation = brevitas_activations[activation]
    return brevitas_activation(bit_width=bit_width, scaling_per_channel=scaling_per_channel, quant_type=QUANT_TYPE,
                               scaling_impl_type=ACT_SCALING_IMPL_TYPE, scaling_min_val=SCALING_MIN_VAL, restrict_scaling_type=ACT_RESTRICT_SCALING_TYPE,
                               max_val=absolute_act_val, per_channel_broadcastable_shape=(1, channels, 1),
                               scaling_stats_permute_dims=(1, 0, 2), return_quant_tensor=False)


def make_quantconv1d(feat_in, classes, kernel_size, bit_width, scaling_per_channel, bias,
                     stride=1, padding=0, dilation=1, groups=1):

    return quant_nn.QuantConv1d(in_channels=feat_in, out_channels=classes, kernel_size=kernel_size,
                                stride=stride, padding=padding, dilation=dilation, groups=groups,
                                bias=bias,
                                weight_bit_width=bit_width,
                                weight_scaling_stats_op=WEIGHT_SCALING_STATS_OP,
                                weight_scaling_per_output_channel=scaling_per_channel,
                                weight_quant_type=QUANT_TYPE,
                                weight_narrow_range=WEIGHT_NARROW_RANGE,
                                weight_scaling_impl_type=WEIGHT_SCALING_IMPL_TYPE,
                                weight_scaling_min_val=SCALING_MIN_VAL,
                                bias_bit_width=bit_width,
                                bias_quant_type=QUANT_TYPE_BIAS,
                                bias_narrow_range=BIAS_CONFIGS,
                                compute_output_scale=BIAS_CONFIGS,
                                compute_output_bit_width=BIAS_CONFIGS,
                                return_quant_tensor=False)



# Copyright (c) 2018-     Xilinx, Inc              (Alessandro Pappalardo)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from warnings import warn
from abc import ABCMeta, abstractmethod
from typing import Type, Union, Optional

from torch.nn import Module
from brevitas.inject import ExtendedInjector, Injector
from brevitas.quant import NoneActQuant
from brevitas.proxy.runtime_quant import ActQuantProxyFromInjector
from brevitas.proxy.runtime_quant import ActQuantProxyProtocol

from .base import QuantProxyMixin


ActQuantType = Union[ActQuantProxyProtocol, Type[Injector], Type[ExtendedInjector]]


class QuantInputMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, input_quant: Optional[ActQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=input_quant,
            proxy_from_injector_impl=ActQuantProxyFromInjector,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='input_',
            kwargs_prefix='input_',
            input_act_impl=None,
            input_passthrough_act=True,
            **kwargs)

    @property
    def is_input_quant_enabled(self):
        return self.input_quant.is_quant_enabled

    @property
    def is_quant_input_narrow_range(self): # TODO make abstract once narrow range can be cached
        return self.input_quant.is_narrow_range

    @property
    @abstractmethod
    def is_quant_input_signed(self):
        pass

    @abstractmethod
    def quant_input_scale(self):
        pass

    @abstractmethod
    def quant_input_zero_point(self):
        pass

    @abstractmethod
    def quant_input_bit_width(self):
        pass


class QuantOutputMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(self, output_quant: Optional[ActQuantType], **kwargs):
        QuantProxyMixin.__init__(
            self,
            quant=output_quant,
            proxy_from_injector_impl=ActQuantProxyFromInjector,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='output_',
            kwargs_prefix='output_',
            output_act_impl=None,
            output_passthrough_act=True,
            **kwargs)

    @property
    def is_output_quant_enabled(self):
        return self.output_quant.is_quant_enabled

    @property
    def is_quant_output_narrow_range(self):  # TODO make abstract once narrow range can be cached
        return self.output_quant.is_narrow_range

    @property
    @abstractmethod
    def is_quant_output_signed(self):
        pass

    @abstractmethod
    def quant_output_scale(self):
        pass

    @abstractmethod
    def quant_output_zero_point(self):
        pass

    @abstractmethod
    def quant_output_bit_width(self):
        pass


class QuantNonLinearActMixin(QuantProxyMixin):
    __metaclass__ = ABCMeta

    def __init__(
            self,
            act_impl: Optional[Type[Module]],
            passthrough_act: bool,
            act_quant: Optional[ActQuantType],
            **kwargs):
        QuantProxyMixin.__init__(
            self,
            act_impl=act_impl,
            passthrough_act=passthrough_act,
            quant=act_quant,
            proxy_from_injector_impl=ActQuantProxyFromInjector,
            proxy_protocol=ActQuantProxyProtocol,
            none_quant_injector=NoneActQuant,
            proxy_prefix='act_',
            kwargs_prefix='',
            **kwargs)

    @property
    def is_act_quant_enabled(self):
        return self.act_quant.is_quant_enabled

    @property
    def is_quant_act_narrow_range(self):  # TODO make abstract once narrow range can be cached
        return self.act_quant.is_narrow_range

    @property
    @abstractmethod
    def is_quant_act_signed(self):
        pass

    @abstractmethod
    def quant_act_scale(self):
        pass

    @abstractmethod
    def quant_act_zero_point(self):
        pass

    @abstractmethod
    def quant_act_bit_width(self):
        pass





from abc import ABC

from torch import Tensor

from brevitas.export.onnx.handler import ONNXBaseHandler
from brevitas.proxy import WeightQuantProxyFromInjector
from brevitas.proxy import BiasQuantProxyFromInjector
from brevitas.proxy import ActQuantProxyFromInjector
from brevitas.proxy.runtime_quant import TruncQuantProxyFromInjector

from .function import QuantPlaceholderFunction, TruncPlaceholderFunction


class StaticQuantProxyHandler(ONNXBaseHandler, ABC):

    def prepare_for_export(self, module):
        if module.is_quant_enabled:
            self.symbolic_kwargs = {
                'scale': module.scale(),
                'zero_point': module.zero_point(),
                'bit_width': module.bit_width(),
                'narrow_range': module.is_narrow_range,
                'signed': module.is_signed}

    def symbolic_execution(self, x: Tensor):
        x = QuantPlaceholderFunction.apply(x, *self.symbolic_kwargs.values())
        scale = self.symbolic_kwargs['scale']
        zero_point = self.symbolic_kwargs['zero_point']
        bit_width = self.symbolic_kwargs['bit_width']
        return x, scale, zero_point, bit_width


class WeightQuantProxyHandler(StaticQuantProxyHandler):
    handled_layer = WeightQuantProxyFromInjector


class ActQuantProxyHandler(StaticQuantProxyHandler):
    handled_layer = ActQuantProxyFromInjector


class BiasQuantProxyHandler(StaticQuantProxyHandler):
    handled_layer = BiasQuantProxyFromInjector

    def symbolic_execution(self, x: Tensor, input_scale=None, input_bit_width=None):
        scale = self.symbolic_kwargs.pop('scale')
        bit_width = self.symbolic_kwargs.pop('bit_width')
        zero_point = self.symbolic_kwargs.pop('zero_point')
        if scale is None:
            assert input_scale is not None, 'Input scale required for bias export'
            scale = input_scale
        if bit_width is None:
            assert input_bit_width is not None, 'Input bit_width required for bias export'
            bit_width = input_bit_width
        x = QuantPlaceholderFunction.apply(
            x, scale, zero_point, bit_width, *self.symbolic_kwargs.values())
        return x, scale, zero_point, bit_width


class TruncQuantProxyHandler(ONNXBaseHandler):
    handled_layer = TruncQuantProxyFromInjector

    def prepare_for_export(self, module: TruncQuantProxyFromInjector):
        self.symbolic_kwargs = {
                'bit_width': module.bit_width()}

    def symbolic_execution(self, x: Tensor, scale: Tensor, zero_point: Tensor, bit_width: Tensor):
        x = TruncPlaceholderFunction.apply(
            x, scale, zero_point, *self.symbolic_kwargs.values())
        return x, scale, zero_point, self.symbolic_kwargs['bit_width']


# Copyright (c) 2018-     Xilinx, Inc              (Alessandro Pappalardo)
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# 3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
#    NEC Laboratories America and IDIAP Research Institute nor the names
#    of its contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Union, Type, List, Optional

from torch import Tensor
from torch.nn import Module

from brevitas.quant_tensor import QuantTensor
from brevitas.inject.defaults import Int8ActPerTensorFloat
from .quant_layer import QuantInputOutputLayer, ActQuantType


class QuantEltwiseAdd(QuantInputOutputLayer, Module):

    def __init__(
            self,
            input_quant: Optional[ActQuantType] = Int8ActPerTensorFloat,
            output_quant: Optional[ActQuantType] = Int8ActPerTensorFloat,
            tie_input_output_quant = False,
            return_quant_tensor: bool = False,
            **kwargs) -> None:
        Module.__init__(self)
        QuantInputOutputLayer.__init__(
            self,
            input_quant,
            output_quant,
            tie_input_output_quant,
            return_quant_tensor,
            **kwargs)

    @property
    def channelwise_separable(self) -> bool:
        return True

    def forward(
            self,
            input: Union[Tensor, QuantTensor],
            other: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        input = self.unpack_input(input)
        other = self.unpack_input(other)
        if self.export_mode:
            assert self.cache_quant_io_metadata_only, "Can't cache multiple inputs"
            out = self.export_handler(inp=input.value, other=other.value)
            self._set_global_is_quant_layer(False)
            return out
        quant_input = self.input_quant(input)
        quant_other = self.input_quant(other)
        output = quant_input + quant_other
        quant_output = self.output_quant(output)
        return self.pack_output(quant_output)


class QuantCat(QuantInputOutputLayer, Module):

    def __init__(
            self,
            input_quant: Optional[ActQuantType] = Int8ActPerTensorFloat,
            output_quant: Optional[ActQuantType] = Int8ActPerTensorFloat,
            tie_input_output_quant: bool = False,
            return_quant_tensor: bool = False,
            **kwargs):
        Module.__init__(self)
        QuantInputOutputLayer.__init__(
            self,
            input_quant,
            output_quant,
            tie_input_output_quant,
            return_quant_tensor,
            **kwargs)

    @property
    def channelwise_separable(self) -> bool:
        return True

    def forward(
            self,
            tensor_list: Union[List[Tensor], List[QuantTensor]],
            dim: int = 1) -> Union[Tensor, QuantTensor]:
        quant_tensor_list = [self.unpack_input(t) for t in tensor_list]
        # shortcut execution through the export impl during export
        if self.export_mode:
            out = self.export_handler([qt.value for qt in quant_tensor_list])
            self._set_global_is_quant_layer(False)
            return out
        quant_tensor_list = [self.input_quant(qt) for qt in quant_tensor_list]
        # trigger an assert if scale factors and bit widths are None or different
        output = QuantTensor.cat(quant_tensor_list, dim=dim)
        quant_output = self.output_quant(output)
        return self.pack_output(quant_output)






import torch


def module_class_name(m: torch.nn.Module):
    module = m.__class__.__module__
    if module is None or module == str.__class__.__module__:
        full_name = m.__class__.__name__
    else:
        full_name = module + '.' + m.__class__.__name__
    return full_name


def _inner_flatten(container):
    for i in container:
        if isinstance(i, (list, tuple)):
            for j in _inner_flatten(i):
                yield j
        else:
            yield i


def flatten(container):
    if isinstance(container, list):
        return list(_inner_flatten(container))
    elif isinstance(container, tuple):
        return tuple(_inner_flatten(container))
    else:
        return container


# Adapted from https://github.com/NVIDIA/NeMo/blob/r0.9/collections/nemo_asr/
# Copyright (C) 2020 Xilinx (Giuseppe Franco)
# Copyright (C) 2019 NVIDIA CORPORATION.
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from .metrics import word_error_rate


def __ctc_decoder_predictions_tensor(tensor, labels):
    """
    Decodes a sequence of labels to words
    """
    blank_id = len(labels)
    hypotheses = []
    labels_map = dict([(i, labels[i]) for i in range(len(labels))])
    prediction_cpu_tensor = tensor.long().cpu()
    # iterate over batch
    for ind in range(prediction_cpu_tensor.shape[0]):
        # print(prediction)
        prediction = prediction_cpu_tensor[ind].numpy().tolist()
        # CTC decoding procedure
        decoded_prediction = []
        previous = len(labels)  # id of a blank symbol
        for p in prediction:
            if (p != previous or previous == blank_id) and p != blank_id:
                decoded_prediction.append(p)
            previous = p
        hypothesis = ''.join([labels_map[c] for c in decoded_prediction])
        hypotheses.append(hypothesis)
    return hypotheses


def monitor_asr_train_progress(tensors: list,
                               labels: list,
                               eval_metric='WER',
                               tb_logger=None,
                               logger=None):
    """
    Takes output of greedy ctc decoder and performs ctc decoding algorithm to
    remove duplicates and special symbol. Prints sample to screen, computes
    and logs AVG WER to console and (optionally) Tensorboard
    Args:
      tensors: A list of 3 tensors (predictions, targets, target_lengths)
      labels: A list of labels
      eval_metric: An optional string from 'WER', 'CER'. Defaults to 'WER'.
      tb_logger: Tensorboard logging object
      logger:
    Returns:
      None
    """
    references = []

    labels_map = dict([(i, labels[i]) for i in range(len(labels))])
    with torch.no_grad():
        # prediction_cpu_tensor = tensors[0].long().cpu()
        targets_cpu_tensor = tensors[2].long().cpu()
        tgt_lenths_cpu_tensor = tensors[3].long().cpu()

        # iterate over batch
        for ind in range(targets_cpu_tensor.shape[0]):
            tgt_len = tgt_lenths_cpu_tensor[ind].item()
            target = targets_cpu_tensor[ind][:tgt_len].numpy().tolist()
            reference = ''.join([labels_map[c] for c in target])
            references.append(reference)
        hypotheses = __ctc_decoder_predictions_tensor(
            tensors[1], labels=labels)

    eval_metric = eval_metric.upper()
    if eval_metric not in {'WER', 'CER'}:
        raise ValueError('eval_metric must be \'WER\' or \'CER\'')
    use_cer = True if eval_metric == 'CER' else False

    tag = f'training_batch_{eval_metric}'
    wer = word_error_rate(hypotheses, references, use_cer=use_cer)
    if tb_logger is not None:
        tb_logger.add_scalar(tag, wer)
    if logger:
        logger.info(f'Loss: {tensors[0]}')
        logger.info(f'{tag}: {wer*100 : 5.2f}%')
        logger.info(f'Prediction: {hypotheses[0]}')
        logger.info(f'Reference: {references[0]}')
    else:
        print(f'Loss: {tensors[0]}')
        print(f'{tag}: {wer*100 : 5.2f}%')
        print(f'Prediction: {hypotheses[0]}')
        print(f'Reference: {references[0]}')


def __gather_losses(losses_list: list) -> list:
    return [torch.mean(torch.stack(losses_list))]


def __gather_predictions(predictions_list: list, labels: list) -> list:
    results = []
    for prediction in predictions_list:
        results += __ctc_decoder_predictions_tensor(prediction, labels=labels)
    return results


def __gather_transcripts(transcript_list: list, transcript_len_list: list,
                         labels: list) -> list:
    results = []
    labels_map = dict([(i, labels[i]) for i in range(len(labels))])
    # iterate over workers
    for t, ln in zip(transcript_list, transcript_len_list):
        # iterate over batch
        t_lc = t.long().cpu()
        ln_lc = ln.long().cpu()
        for ind in range(t.shape[0]):
            tgt_len = ln_lc[ind].item()
            target = t_lc[ind][:tgt_len].numpy().tolist()
            reference = ''.join([labels_map[c] for c in target])
            results.append(reference)
    return results


def process_evaluation_batch(tensors: dict, global_vars: dict, labels: list):
    """
    Creates a dictionary holding the results from a batch of audio
    """
    if 'EvalLoss' not in global_vars.keys():
        global_vars['EvalLoss'] = []
    if 'predictions' not in global_vars.keys():
        global_vars['predictions'] = []
    if 'transcripts' not in global_vars.keys():
        global_vars['transcripts'] = []
    if 'logits' not in global_vars.keys():
        global_vars['logits'] = []
    # if not 'transcript_lengths' in global_vars.keys():
    #  global_vars['transcript_lengths'] = []
    for kv, v in tensors.items():
        if kv.startswith('loss'):
            global_vars['EvalLoss'] += __gather_losses(v)
        elif kv.startswith('predictions'):
            global_vars['predictions'] += __gather_predictions(
                v, labels=labels)
        elif kv.startswith('transcript_length'):
            transcript_len_list = v
        elif kv.startswith('transcript'):
            transcript_list = v
        elif kv.startswith('output'):
            global_vars['logits'] += v

    global_vars['transcripts'] += __gather_transcripts(transcript_list,
                                                       transcript_len_list,
                                                       labels=labels)


def process_evaluation_epoch(global_vars: dict,
                             eval_metric='WER',
                             tag=None,
                             logger=None):
    """
    Calculates the aggregated loss and WER across the entire evaluation dataset
    """
    eloss = torch.mean(torch.stack(global_vars['EvalLoss'])).item()
    hypotheses = global_vars['predictions']
    references = global_vars['transcripts']

    eval_metric = eval_metric.upper()
    if eval_metric not in {'WER', 'CER'}:
        raise ValueError('eval_metric must be \'WER\' or \'CER\'')
    use_cer = True if eval_metric == 'CER' else False

    wer = word_error_rate(hypotheses=hypotheses,
                          references=references,
                          use_cer=use_cer)

    if tag is None:
        if logger:
            logger.info(f"==========>>>>>>Evaluation Loss: {eloss}")
            logger.info(f"==========>>>>>>Evaluation {eval_metric}: "
                        f"{wer*100 : 5.2f}%")
        else:
            print(f"==========>>>>>>Evaluation Loss: {eloss}")
            print(f"==========>>>>>>Evaluation {eval_metric}: "
                  f"{wer*100 : 5.2f}%")
        return {"Evaluation_Loss": eloss, f"Evaluation_{eval_metric}": wer}
    else:
        if logger:
            logger.info(f"==========>>>>>>Evaluation Loss {tag}: {eloss}")
            logger.info(f"==========>>>>>>Evaluation {eval_metric} {tag}: "
                        f"{wer*100 : 5.2f}%")
        else:
            print(f"==========>>>>>>Evaluation Loss {tag}: {eloss}")
            print(f"==========>>>>>>Evaluation {eval_metric} {tag}:"
                  f" {wer*100 : 5.2f}%")
        return {f"Evaluation_Loss_{tag}": eloss,
                f"Evaluation_{eval_metric}_{tag}": wer}


def post_process_predictions(predictions, labels):
    return __gather_predictions(predictions, labels=labels)


def post_process_transcripts(
        transcript_list, transcript_len_list, labels):
    return __gather_transcripts(transcript_list,
                                transcript_len_list,
                                labels=labels)



from abc import ABC
from typing import Optional, Union

import torch
from torch import Tensor

from brevitas.nn import QuantLinear, QuantConv2d, QuantConv1d
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from brevitas.export.onnx.handler import Kernel2dApplHandler, Kernel1dApplHandler
from .base import FINNQuantIOHandler
from ..function.parameter import QuantizedLinearPlaceholderFunction
from ..function.parameter import QuantizedConvNdPlaceholderFunction


QuantConvNd = Union[QuantConv1d, QuantConv2d]


class FINNQuantWBIOLHandler(FINNQuantIOHandler, ABC):

    @staticmethod
    def sanity_check(module: QuantWBIOL):
        assert module.is_weight_quant_enabled
        assert not module.is_input_quant_enabled
        assert not module.is_output_quant_enabled

    @staticmethod
    def quant_weight_type(module: QuantWBIOL):
        bit_width = int(module.quant_weight_bit_width().item())
        if bit_width == 1:
            return "BIPOLAR"
        else:
            return f"INT{bit_width}"

    @staticmethod
    def int_weight(module: QuantConvNd):
        return module.int_weight(float_datatype=True).detach()

    @staticmethod
    def int_weight_transposed(module: QuantWBIOL):
        return torch.t(module.int_weight(float_datatype=True)).detach()

    @staticmethod
    def function_scaling(
            module: QuantWBIOL,
            quant_input_scale,
            quant_weight_scale,
            quant_output_scale):
        # if bias quant is disabled, whether bias is enabled or not, I just need to
        # pass quant_weight_scale as the function-level scale_factor
        if not module.is_bias_quant_enabled:
            in_scale, scale_factor = None, quant_weight_scale
        else:
            in_scale, scale_factor = quant_input_scale, quant_output_scale
        return in_scale, scale_factor

    @staticmethod
    def maybe_quant_bias(
            module: QuantWBIOL,
            quant_weight_scale: Tensor,
            quant_output_scale: Optional[Union[Tensor, float]],
            quant_bit_width: Optional[Tensor]):
        if module.bias is None:
            return None
        elif module.bias is not None and not module.is_bias_quant_enabled:
            bias = torch.t(module.bias.type(torch.FloatTensor)).detach()
            # account for both scalar tensors and array tensors with 1 elements
            if len(quant_weight_scale.shape) > 0 and len(quant_weight_scale) > 1:
                quant_weight_scale = quant_weight_scale.reshape(bias.shape)
            # divide by weight scale as add is before mul
            bias /= quant_weight_scale
            return bias
        else:  # bias quant enabled
            assert quant_output_scale, 'Quant bias export requires caching of the output scale'
            if not isinstance(quant_output_scale, Tensor):  # item might have been called
                quant_output_scale = torch.tensor(quant_output_scale)
            quant_bias = module.bias_quant(module.bias, quant_output_scale, quant_bit_width)
            quant_bias = torch.t(quant_bias.value.type(torch.FloatTensor)).detach()
            quant_bias /= quant_output_scale
            quant_bias = torch.round(quant_bias)
            return quant_bias


class FINNQuantLinearHandler(FINNQuantWBIOLHandler):
    handled_layer = QuantLinear

    @staticmethod
    def quant_weight_scale(module: QuantWBIOL):
        return torch.t(module.quant_weight_scale().type(torch.FloatTensor)).detach()

    @staticmethod
    def quant_output_shape(module: QuantLinear):
        shape = FINNQuantWBIOLHandler.quant_output_shape(module)
        if shape is None:
            return (1, module.out_features)
        else:
            return shape

    def prepare_for_export(self, module):
        self.sanity_check(module)
        quant_input_scale = self.quant_input_scale(module)
        quant_weight_scale = self.quant_weight_scale(module)
        quant_output_scale = self.quant_output_scale(module)
        quant_output_bit_width_tensor = self.quant_output_bit_width_tensor(module)
        maybe_quant_bias = self.maybe_quant_bias(
            module,
            quant_weight_scale,
            quant_output_scale,
            quant_output_bit_width_tensor)
        in_scale, scale_factor = self.function_scaling(
            module, quant_input_scale, quant_weight_scale, quant_output_scale)
        self.symbolic_kwargs = {
            'Wt': self.int_weight_transposed(module),
            'scale_factor': scale_factor,
            'w_qnt_type': self.quant_weight_type(module),
            'out_shape': self.quant_output_shape(module),
            'bias': maybe_quant_bias,
            'in_scale': in_scale,
            'in_qnt_type': self.quant_input_type(module)}

    def symbolic_execution(self, inp: Tensor):
        ret = QuantizedLinearPlaceholderFunction.apply(inp, *self.symbolic_kwargs.values())
        return ret


class FINNQuantConvNdHandler(FINNQuantWBIOLHandler, ABC):

    @staticmethod
    def quant_output_shape(module: QuantConvNd):
        shape = FINNQuantWBIOLHandler.quant_output_shape(module)
        if shape is None:
            raise RuntimeError("Caching of output shapes is required to export QuantConvNd")
        return shape

    @staticmethod
    def maybe_quant_bias(
            module: QuantWBIOL,
            quant_weight_scale: Tensor,
            quant_output_scale: Optional[Union[Tensor, float]],
            quant_bit_width: Optional[Tensor]):
        quant_bias = FINNQuantWBIOLHandler.maybe_quant_bias(
            module, quant_weight_scale, quant_output_scale, quant_bit_width)
        if quant_bias is not None:
            quant_bias_shape = [1] * len(module.weight.shape)
            quant_bias_shape[1] = -1 # shape should broadcast with activations along channel dim
            quant_bias = quant_bias.view(quant_bias_shape)
        return quant_bias

    def prepare_for_export(self, module: QuantConvNd):
        self.sanity_check(module)
        quant_input_scale = self.quant_input_scale(module)
        quant_weight_scale = self.quant_weight_scale(module)
        quant_output_scale = self.quant_output_scale(module)
        quant_output_bit_width_tensor = self.quant_output_bit_width_tensor(module)
        maybe_quant_bias = self.maybe_quant_bias(
            module,
            quant_weight_scale,
            quant_output_scale,
            quant_output_bit_width_tensor)
        in_scale, scale_factor = self.function_scaling(
            module, quant_input_scale, quant_weight_scale, quant_output_scale)
        self.symbolic_kwargs = {
            'W': self.int_weight(module),
            'scale_factor': scale_factor,
            'w_qnt_type': self.quant_weight_type(module),
            'out_shape': self.quant_output_shape(module),
            'pads': self.padding(module),
            'strides': self.stride(module),
            'bias': maybe_quant_bias,
            'in_scale': in_scale,
            'in_qnt_type': self.quant_input_type(module),
            'kernel_shape': list(module.kernel_size),
            'groups': module.groups,
            'dilations': self.dilation(module)}

    def symbolic_execution(self, inp: Tensor):
        ret = QuantizedConvNdPlaceholderFunction.apply(inp, *self.symbolic_kwargs.values())
        return ret


class FINNQuantConv1dHandler(FINNQuantConvNdHandler, Kernel1dApplHandler):
    handled_layer = QuantConv1d

    @staticmethod
    def quant_weight_scale(module: QuantConv1d):
        quant_weight_scale = module.quant_weight_scale().type(torch.FloatTensor).detach()
        if len(quant_weight_scale.shape) == 3:
            quant_weight_scale = quant_weight_scale.view(1, -1, 1)
        return quant_weight_scale


class FINNQuantConv2dHandler(FINNQuantConvNdHandler, Kernel2dApplHandler):
    handled_layer = QuantConv2d

    @staticmethod
    def quant_weight_scale(module: QuantConv2d):
        quant_weight_scale = module.quant_weight_scale().type(torch.FloatTensor).detach()
        if len(quant_weight_scale.shape) == 4:
            quant_weight_scale = quant_weight_scale.view(1, -1, 1, 1)
        return quant_weight_scale



# Audio dataset and corresponding functions taken from Patter  https://github.com/ryanleary/patter
# Adapted from https://github.com/NVIDIA/NeMo/tree/r0.9/collections/nemo_asr
# MIT License
# Copyright (c) 2020 Xilinx (Giuseppe Franco)
# Copyright (c) 2019 NVIDIA Corporation
# Copyright (c) 2018 Ryan Leary
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import random

import librosa
from scipy import signal

from .manifest import ManifestEN
from .segment import AudioSegment


class Perturbation(object):
    def max_augmentation_length(self, length):
        return length

    def perturb(self, data):
        raise NotImplementedError


class SpeedPerturbation(Perturbation):
    def __init__(self, min_speed_rate=0.85, max_speed_rate=1.15, rng=None):
        self._min_rate = min_speed_rate
        self._max_rate = max_speed_rate
        self._rng = random.Random() if rng is None else rng

    def max_augmentation_length(self, length):
        return length * self._max_rate

    def perturb(self, data):
        speed_rate = self._rng.uniform(self._min_rate, self._max_rate)
        if speed_rate <= 0:
            raise ValueError("speed_rate should be greater than zero.")
        # print("DEBUG: speed:", speed_rate)
        data._samples = librosa.effects.time_stretch(data._samples, speed_rate)


class GainPerturbation(Perturbation):
    def __init__(self, min_gain_dbfs=-10, max_gain_dbfs=10, rng=None):
        self._min_gain_dbfs = min_gain_dbfs
        self._max_gain_dbfs = max_gain_dbfs
        self._rng = random.Random() if rng is None else rng

    def perturb(self, data):
        gain = self._rng.uniform(self._min_gain_dbfs, self._max_gain_dbfs)
        # print("DEBUG: gain:", gain)
        data._samples = data._samples * (10. ** (gain / 20.))


class ImpulsePerturbation(Perturbation):
    def __init__(self, manifest_path=None, rng=None):
        self._manifest = ManifestEN(manifest_path)
        self._rng = random.Random() if rng is None else rng

    def perturb(self, data):
        impulse_record = self._rng.sample(self._manifest.data, 1)[0]
        impulse = AudioSegment.from_file(impulse_record['audio_filepath'],
                                         target_sr=data.sample_rate)
        # print("DEBUG: impulse:", impulse_record['audio_filepath'])
        data._samples = signal.fftconvolve(
            data.samples, impulse.samples, "full")


class ShiftPerturbation(Perturbation):
    def __init__(self, min_shift_ms=-5.0, max_shift_ms=5.0, rng=None):
        self._min_shift_ms = min_shift_ms
        self._max_shift_ms = max_shift_ms
        self._rng = random.Random() if rng is None else rng

    def perturb(self, data):
        shift_ms = self._rng.uniform(self._min_shift_ms, self._max_shift_ms)
        if abs(shift_ms) / 1000 > data.duration:
            # TODO: do something smarter than just ignore this condition
            return
        shift_samples = int(shift_ms * data.sample_rate // 1000)
        # print("DEBUG: shift:", shift_samples)
        if shift_samples < 0:
            data._samples[-shift_samples:] = data._samples[:shift_samples]
            data._samples[:-shift_samples] = 0
        elif shift_samples > 0:
            data._samples[:-shift_samples] = data._samples[shift_samples:]
            data._samples[-shift_samples:] = 0


class NoisePerturbation(Perturbation):
    def __init__(self, manifest_path=None, min_snr_db=40, max_snr_db=50,
                 max_gain_db=300.0, rng=None):
        self._manifest = ManifestEN(manifest_path)
        self._rng = random.Random() if rng is None else rng
        self._min_snr_db = min_snr_db
        self._max_snr_db = max_snr_db
        self._max_gain_db = max_gain_db

    def perturb(self, data):
        snr_db = self._rng.uniform(self._min_snr_db, self._max_snr_db)
        noise_record = self._rng.sample(self._manifest.data, 1)[0]
        noise = AudioSegment.from_file(noise_record['audio_filepath'],
                                       target_sr=data.sample_rate)
        noise_gain_db = min(data.rms_db - noise.rms_db - snr_db,
                            self._max_gain_db)
        # print("DEBUG: noise:", snr_db, noise_gain_db, noise_record[
        # 'audio_filepath'])

        # calculate noise segment to use
        start_time = self._rng.uniform(0.0, noise.duration - data.duration)
        noise.subsegment(start_time=start_time,
                         end_time=start_time + data.duration)

        # adjust gain for snr purposes and superimpose
        noise.gain_db(noise_gain_db)
        data._samples = data._samples + noise.samples


perturbation_types = {
    "speed": SpeedPerturbation,
    "gain": GainPerturbation,
    "impulse": ImpulsePerturbation,
    "shift": ShiftPerturbation,
    "noise": NoisePerturbation
}


class AudioAugmentor(object):
    def __init__(self, perturbations=None, rng=None):
        self._rng = random.Random() if rng is None else rng
        self._pipeline = perturbations if perturbations is not None else []

    def perturb(self, segment):
        for (prob, p) in self._pipeline:
            if self._rng.random() < prob:
                p.perturb(segment)
        return

    def max_augmentation_length(self, length):
        newlen = length
        for (prob, p) in self._pipeline:
            newlen = p.max_augmentation_length(newlen)
        return newlen

    @classmethod
    def from_config(cls, config):
        ptbs = []
        for p in config:
            if p['aug_type'] not in perturbation_types:
                print(p['aug_type'], "perturbation not known. Skipping.")
                continue
            perturbation = perturbation_types[p['aug_type']]
            ptbs.append((p['prob'], perturbation(**p['cfg'])))
        return cls(perturbations=ptbs)



from inspect import signature
from abc import abstractmethod, ABC
from copy import deepcopy

import torch
from torch.nn import Module
from torch.nn import Conv2d, Conv1d, Linear

from brevitas import nn as qnn
from brevitas.quant_tensor import QuantTensor
from brevitas.nn.quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from brevitas.nn.utils import merge_bn
from .module import CodegenModule, _replace_module, Instruction, FnType, _set_module


def _signature_keys(module_class):
    return signature(module_class).parameters.keys()


class Rewriter(ABC):
    _module_prefix_id_map = {}

    @abstractmethod
    def apply(self, model: CodegenModule) -> CodegenModule:
        pass

    def module_id(self, name):
        if name in self._module_prefix_id_map:
            self._module_prefix_id_map[name] += 1
        else:
            self._module_prefix_id_map[name] = 0
        return self._module_prefix_id_map[name]


class InstRewriter(Rewriter, ABC):

    @abstractmethod
    def rewrite_inst(self, inst: Instruction, model: CodegenModule):
        pass

    def apply(self, model: CodegenModule):
        for inst in model.schedule:
            self.rewrite_inst(inst, model)
        return model


class ModuleToModuleRewriter(Rewriter, ABC):

    def __init__(
            self,
            old_module_class,
            new_module_class,
            old_module_cond=lambda old_module: True, **kwargs):
        super().__init__()
        self.old_module_class = old_module_class
        self.new_module_class = new_module_class
        self.new_module_kwargs = kwargs
        self.old_module_cond = old_module_cond

    def map_origin_vars(self, vars: dict):
        return {k: v is not None if k == 'bias' else v for k, v in vars.items()}

    def module_attributes(self, module):
        attrs = vars(module)
        # workaround since bias doesn't show up on vars of Linear
        if hasattr(module, 'bias'):
            attrs['bias'] = module.bias
        return attrs

    def _init_new_module(self, old_module: Module):
        # get attributes of original module
        new_kwargs = self.module_attributes(old_module)
        # transforms attribute of original module, e.g. bias Parameter -> bool
        new_kwargs = self.map_origin_vars(new_kwargs)
        # restrict to only values that are in the init of the new module
        new_module_signature_keys = _signature_keys(self.new_module_class)
        new_kwargs = {k: v for k, v in new_kwargs.items() if k in new_module_signature_keys}
        # update with kwargs passed to the rewriter
        new_kwargs.update(self.new_module_kwargs)
        # init the new module
        new_module = self.new_module_class(**new_kwargs)
        return new_module

    def _rewrite_schedule(self, old_module_use_list, new_module):
        # replace old module references in the compute graph
        for inst in old_module_use_list:
            inst.fn = new_module

    def _rewrite_model(self, model, old_new_module_dict):
        for old_module, new_module in old_new_module_dict.items():
            _replace_module(model, old_module, new_module)
            new_module.load_state_dict(old_module.state_dict())

    def apply(self, model: CodegenModule):
        old_new_module_dict = {}
        for old_module in model.modules():
            # check for equality, not inheritance
            if type(old_module) == self.old_module_class and self.old_module_cond(old_module):
                # init the new module based on the old one
                new_module = self._init_new_module(old_module)
                # rewrite compute schedule
                old_module_use_list = model.module_use_dict()[old_module]
                self._rewrite_schedule(old_module_use_list, new_module)
                # register modules pair to be replaced
                old_new_module_dict[old_module] = new_module
        # replace all pairs registered
        self._rewrite_model(model, old_new_module_dict)
        return model


class MergeBatchNorm2d(Rewriter):

    merge_layers = [QuantWBIOL, Conv2d, Conv1d, Linear]

    def apply(self, model: CodegenModule):
        for inst in model.schedule:
            if type(inst.fn) == torch.nn.BatchNorm2d:
                num_args = len(inst.input_args_list)
                num_kwargs = len(inst.input_kwargs_dict.items())
                assert num_args + num_kwargs == 1
                try:
                    bn_input_index = inst.input_args_list[0]
                except:
                    bn_input_index = list(inst.input_kwargs_dict.values())[0]
                inst.merge_into_predecessor = None
                inst.fn.merged_inst = []
                # check that the predecessor doesn't feed into other layers
                can_merge = True
                for predecessor in model.schedule:
                    if predecessor.fn is not inst.fn:
                        for pii in predecessor.input_args_list:
                            for iii in inst.input_args_list:
                                can_merge &= pii != iii
                # identify predecessor and merge bn into it
                for predecessor in model.schedule:
                    if predecessor.output_index == bn_input_index:
                        if any([isinstance(predecessor.fn, l) for l in self.merge_layers]):
                            merge_bn(predecessor.fn, inst.fn)
                            inst.merge_into_predecessor = predecessor
                            inst.fn.merged_inst.append(True)
                        else:
                            inst.fn.merged_inst.append(False)
                # update output index of the predecessor and mark inst as to remove
                if inst.merge_into_predecessor is not None:
                    predecessor = inst.merge_into_predecessor
                    predecessor.output_index = inst.output_index
                    inst.output_index = None
                else:
                    del inst.merge_into_predecessor
        # update schedule to remove merged inst
        model.schedule = [i for i in model.schedule if i.output_index is not None]
        # update module hierarchy to remove merged bn
        for m in list(model.modules()):
            cond = lambda x: not hasattr(x, 'merged_inst') or not all(x.merged_inst)
            m._modules = {name: sm for name, sm in m._modules.items() if cond(sm)}
        # remove leftover metadata on bn that hasn't been merged
        for m in model.modules():
            if hasattr(m, 'merged_inst'):
                del m.merged_inst
        return model


class DuplicateSharedStatelessModule(Rewriter):

    def apply(self, model: CodegenModule):
        for m, use_list in model.module_use_dict().items():
            is_stateful = list(m.parameters()) or list(m.buffers())
            if len(use_list) > 1 and not is_stateful:
                for use_inst in use_list[1:]:
                    dup_m = deepcopy(m)
                    use_inst.fn = dup_m
                    model.add_module(str(use_inst.output_index), dup_m)
        return model


class DisableBreakingReturnQuantTensor(Rewriter):

    def apply(self, model: CodegenModule):
        for inst in model.schedule:
            if hasattr(inst.fn, 'return_quant_tensor') and inst.fn.return_quant_tensor:
                s_list = model.inst_successor_list(inst)
                cond = lambda s: (((s.fn_type == FnType.MODULE or s.fn_type == FnType.SCRIPTMODULE)
                                   and hasattr(s.fn, 'accept_quant_tensor')
                                   or (s.fn_type == FnType.ATTRIBUTE or s.fn_type == FnType.METHOD)
                                   and hasattr(QuantTensor, s.fn))
                                  and not s.fn_type == FnType.FUNCTION)
                if not all([cond(s) for s in s_list]):
                    inst.fn.return_quant_tensor = False
        return model


class CallableToModuleRewriter(Rewriter, ABC):

    def __init__(
            self,
            old_callable,
            new_module_class,
            inst_cond=lambda old_inst: True,
            inst_postprocess=lambda new_inst: new_inst,
            **kwargs):
        super().__init__()
        self.old_callable = old_callable
        self.new_module_class = new_module_class
        self.new_module_kwargs = kwargs
        self.inst_cond = inst_cond
        self.inst_postprocess = inst_postprocess

    @abstractmethod
    def match_inst(self, inst):
        pass

    def new_module_prefix(self, inst):
        module_prefix = inst.prefix + '.' + inst.fn_name
        module_prefix = module_prefix + str(self.module_id(module_prefix))
        return module_prefix

    def update_kwargs(self, inst: Instruction):
        new_module_keys = _signature_keys(self.new_module_class)
        old_fn_kwargs = inst.input_kwargs_dict
        old_fn_kwargs_keys = list(old_fn_kwargs.keys())
        extracted = {k: old_fn_kwargs.pop(k) for k in old_fn_kwargs_keys if k in new_module_keys}
        self.new_module_kwargs.update(extracted)

    def gen_new_module(self, inst, model):
        module_prefix = self.new_module_prefix(inst)
        self.update_kwargs(inst)
        module = self.new_module_class(**self.new_module_kwargs)
        _set_module(model, module, module_prefix)
        return module

    def rewrite_inst(self, inst, model):
        module = self.gen_new_module(inst, model)
        inst.fn = module
        inst.fn_type = FnType.MODULE
        self.inst_postprocess(inst)

    def apply(self, model: CodegenModule) -> CodegenModule:
        for inst in model.schedule:
            if self.match_inst(inst):
                self.rewrite_inst(inst, model)
        return model


class TorchFnToModuleRewriter(CallableToModuleRewriter):

    def match_inst(self, inst) -> bool:
        return (self.inst_cond(inst)
                and inst.fn_type == FnType.FUNCTION
                and inst.fn == self.old_callable)


class TensorMethodToModuleRewriter(CallableToModuleRewriter):

    def match_inst(self, inst) -> bool:
        return (self.inst_cond(inst)
                and inst.fn_type == FnType.METHOD
                and inst.fn == self.old_callable)

    def update_kwargs(self, inst: Instruction):
        super(TensorMethodToModuleRewriter, self).update_kwargs(inst)
        tensor = inst.input_kwargs_dict.pop('self')
        inst.input_args_list = [tensor] + inst.input_args_list


class MeanToAdaptiveAvgPool2d(TensorMethodToModuleRewriter):

    def __init__(self):
        super(MeanToAdaptiveAvgPool2d, self).__init__(
            old_callable='mean',
            new_module_class=qnn.QuantAdaptiveAvgPool2d,
            inst_cond=MeanToAdaptiveAvgPool2d.match_mean_cond,
            inst_postprocess=MeanToAdaptiveAvgPool2d.postprocess_inst,
            output_size=(1, 1))

    @staticmethod
    def match_mean_cond(inst):
        return ([2, 3] in inst.input_args_list
                or (2, 3) in inst.input_args_list
                or 'dim' in inst.input_kwargs_dict
                and (inst.input_kwargs_dict['dim'] == (2, 3)
                     or inst.input_kwargs_dict['dim'] == [2, 3]))

    @staticmethod
    def postprocess_inst(inst):
        if 'dim' in inst.input_kwargs_dict:
            del inst.input_kwargs_dict['dim']
        elif (2, 3) in inst.input_args_list:
            inst.input_args_list.remove((2, 3))
        elif [2, 3] in inst.input_args_list:
            inst.input_args_list.remove([2, 3])
        # TODO this is a hack, a separate view instruction should be generated
        original_forward = inst.fn.forward
        def forward_with_view(*args, **kwargs):
            out = original_forward(*args, **kwargs)
            return out.view(int(out.size(0)), -1)
        inst.fn.forward = forward_with_view


class InplaceMeanToAdaptiveAvgPool2d(TensorMethodToModuleRewriter):

    def __init__(self):
        super(InplaceMeanToAdaptiveAvgPool2d, self).__init__(
            old_callable='mean_',
            new_module_class=qnn.QuantAdaptiveAvgPool2d,
            inst_cond=MeanToAdaptiveAvgPool2d.match_mean_cond,
            inst_postprocess=MeanToAdaptiveAvgPool2d.postprocess_inst,
            output_size=(1, 1))


class RewriterList(Rewriter):

    def __init__(self, rewriter_list):
        self.rewriter_list = rewriter_list

    def apply(self, model: CodegenModule):
        for rewriter in self.rewriter_list:
            rewriter.apply(model)
        return model
